from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

import matplotlib.pyplot as plt

from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error, mean_absolute_error

from spicy_snow.retrieval import retrieval_from_parameters

# import functions for downloading
from spicy_snow.download.sentinel1 import s1_img_search, hyp3_pipeline, download_hyp3, combine_s1_images
from spicy_snow.download.forest_cover import download_fcf
from spicy_snow.download.snow_cover import download_snow_cover

# import functions for pre-processing
from spicy_snow.processing.s1_preprocessing import merge_partial_s1_images, s1_orbit_averaging,\
s1_clip_outliers, subset_s1_images, ims_water_mask, s1_incidence_angle_masking, merge_s1_subsets

# import the functions for snow_index calculation
from spicy_snow.processing.snow_index import calc_delta_VV, calc_delta_cross_ratio, \
    calc_delta_gamma, clip_delta_gamma_outlier, calc_snow_index, calc_snow_index_to_snow_depth

# import the functions for wet snow flag
from spicy_snow.processing.wet_snow import id_newly_frozen_snow, id_newly_wet_snow, \
    id_wet_negative_si, flag_wet_snow

# setup root logger
from spicy_snow.utils.spicy_logging import setup_logging

# fischer z test
from causallearn.utils.cit import CIT

# random fields
from gstools import SRF
from gstools import vario_estimate_unstructured
from gstools import Exponential

def s1_to_sd(ds, A = 1.5, B = 0.1, C = 0.6):
    # mask out outliers in incidence angle
    ds = s1_incidence_angle_masking(ds)
    
    # subset dataset by flight_dir and platform
    dict_ds = subset_s1_images(ds)

    for subset_name, subset_ds in dict_ds.items():
        # average each orbit to overall mean
        dict_ds[subset_name] = s1_orbit_averaging(subset_ds)
        # clip outlier values of backscatter to overall mean
        dict_ds[subset_name] = s1_clip_outliers(subset_ds)
    
    # recombine subsets
    ds = merge_s1_subsets(dict_ds)

    # calculate confidence interval
    # ds = add_confidence_angle(ds)

    ## Snow Index Steps
    # log.info("Calculating snow index")
    # calculate delta CR and delta VV
    ds = calc_delta_cross_ratio(ds, A = A)
    ds = calc_delta_VV(ds)

    # calculate delta gamma with delta CR and delta VV with FCF
    ds = calc_delta_gamma(ds, B = B)

    # clip outliers of delta gamma
    ds = clip_delta_gamma_outlier(ds)

    # calculate snow_index from delta_gamma
    ds = calc_snow_index(ds, ims_masking = True)

    # convert snow index to snow depth
    ds = calc_snow_index_to_snow_depth(ds, C = C)

    return ds

import warnings
def get_gaussian_stats(da):
    arr = da.values
    warnings.filterwarnings("ignore", message="Mean of empty slice")
    warnings.filterwarnings("ignore", message="Degrees of freedom <= 0 ")
    return arr.size, np.nanmean(arr), np.nanstd(arr), arr.shape

def get_bias(x, y): return np.mean(x - y)

def get_stats(x, y, nrmse = False):
    if type(x) == xr.DataArray: x = x.values.ravel()
    if type(y) == xr.DataArray: y = y.values.ravel()
    if type(x) == list: x = np.array(x)
    if type(y) == list: y = np.array(y)
    idx = (~np.isnan(x)) & (~np.isnan(y))
    x, y = x[idx], y[idx]
    if len(x) < 4 or len(y) < 4:
        if nrmse:
            return np.nan, np.nan, np.nan, np.nan, np.nan
        return np.nan, np.nan, np.nan, np.nan
    r, p = pearsonr(x, y)
    bias = get_bias(x, y)
    mae = mean_absolute_error(x, y)
    rmse = mean_squared_error(x, y, squared = False)

    if nrmse:
        nrmse_value = rmse / np.mean(x)
        return r, b, mae, rmse, nrmse_value
    
    return r, bias, mae, rmse


from scipy.stats import norm
def fischerz(truth, x1, x2):
    idx = (~np.isnan(truth)) & (~np.isnan(x1)) & (~np.isnan(x2))
    truth, x1, x2 = truth[idx], x1[idx], x2[idx]
    n = len(x1)
    cor1 = pearsonr(truth, x1).statistic
    cor2 = pearsonr(truth, x2).statistic
    fischer1 = 0.5*np.log((1+cor1)/(1-cor1))
    fischer2 = 0.5*np.log((1+cor2)/(1-cor2))
    expected_sd = np.sqrt(1/(n-3))
    return 2 * (1 - norm(0, expected_sd).cdf(np.abs(fischer1 - fischer2)))

def optimize(r_ds, A, B, C, im_date):
    # holds rmse and r for optimizing rand dataset
    rand_rmse_ds = xr.DataArray(np.zeros([len(A), len(B), len(C)]), dims = ['A', 'B', 'C'], coords = [A, B, C])
    rand_mae_ds = xr.DataArray(np.zeros([len(A), len(B), len(C)]), dims = ['A', 'B', 'C'], coords = [A, B, C])
    rand_r_ds = xr.DataArray(np.zeros([len(A), len(B), len(C)]), dims = ['A', 'B', 'C'], coords = [A, B, C])

    # optimize random dataset
    for a in A:
        for b in B:
            for c in C:
                r_ds_new = s1_to_sd(r_ds, A = a, B = b, C = c)
                r_sd_new = r_ds_new.where(r_ds_new['ims'] == 4)['snow_depth'].dropna(dim = 'time', how = 'all').sel(time = im_date, method = 'nearest')
                rand_r, rand_b, rand_mae, rand_rmse = get_stats(r_sd_new, r_ds_new['lidar-sd'], nrmse = False)
                rand_rmse_ds.loc[dict(A = a, B = b, C = C)] = rand_rmse
                rand_mae_ds.loc[dict(A = a, B = b, C = C)] = rand_mae
                rand_r_ds.loc[dict(A = a, B = b, C = C)] = rand_r
    
    return rand_rmse_ds, rand_mae_ds, rand_r_ds

def create_unstructured_grid(x_s, y_s):
    x_u, y_u = np.meshgrid(x_s, y_s)
    len_unstruct = len(x_s) * len(y_s)
    x_u = np.reshape(x_u, len_unstruct)
    y_u = np.reshape(y_u, len_unstruct)
    return x_u, y_u

def create_random_field(da):
    random_ds = da.copy(deep = True)
    arr = da.data
    x_u, y_u = create_unstructured_grid(da.x.data, da.y.data)
    bins = np.linspace(20, 1500, 20)
    bin_center, gamma = vario_estimate_unstructured(
            (x_u, y_u), arr.flatten(), bins, sampling_size=3000)
    fit_model = Exponential(dim=2, nugget= 0)
    fit_model.fit_variogram(bin_center, gamma, nugget=True)
    srf = SRF(fit_model, seed=19770928)
    # this transpose appears in all the documentation and is neccessary to make the data fit the original data
    random_ds.data = srf((da.x.data, da.y.data), mesh_type='structured').T

    return random_ds

def convert_to_utm(da): 
    da = da.rio.write_crs('EPSG:4326')
    if 'time' in da.dims:
        da = da.transpose('time', 'y', 'x')
    else:
        da = da.transpose('y', 'x')
    da = da.rio.reproject(da.rio.estimate_utm_crs())

    return da

def normalize_da(da):
    c_min, c_max = da.min(), da.max()
    da = (da - c_min) / (c_max - c_min)

    return da

def reorder_timeseries(dataset, ims = False):
    dataset = dataset.assign_coords(data_time = ('time', dataset.time.data))

    for orbit_name, sub in dataset.groupby('relative_orbit'):
        ts = sub.time
        reorder_ts = ts.to_dataframe().sample(n = len(ts), replace = False).index
        for (old_ts, new_ts) in zip(ts, reorder_ts):
            dataset['s1'].loc[dict(time = old_ts)] = dataset['s1'].sel(time = new_ts)
            dataset['data_time'].loc[dict(time = old_ts)] = pd.to_datetime(new_ts)
            if ims:
                dataset['ims'].loc[dict(time = old_ts)] = dataset['ims'].sel(time = new_ts)
        
    return dataset.sortby('time')

def invert_timerseries(ds):
    random_ds = ds.copy(deep = True)
    for band in ['VV','VH']:
        for ts in ds.time:
            random_ds['s1'].loc[dict(time = ts, band = band)] = np.fliplr(np.flipud(ds['s1'].sel(time = ts, band = band)))
    return random_ds

def make_random_results(out_dir, A, B, C, fp):
    print(fp)
    stem = fp.stem
    # ds, stem = ds_pair
    full_ds = xr.open_dataset(fp)
    full_ds = full_ds.load()
    # full_ds.where(full_ds['lidar-sd'] > 2)

    site_name = stem.replace('_', ' ').replace('Frasier', 'Fraser').split('-')[0]
    im_date = pd.to_datetime(stem.split('_')[-1])

    i = 1
    outfp = out_dir.joinpath(f'{site_name}_{i}.csv')
    while outfp.exists():
        i += 1
        outfp = out_dir.joinpath(f'{site_name}_{i}.csv')

    res = pd.DataFrame(index = pd.MultiIndex.from_tuples(list(zip([], [])), names=["condition", "round"]))

    ds = full_ds.sel(time = im_date, method = 'nearest').copy()

    if site_name not in res.index:
        r, b, mae, rmse = get_stats(ds['snow_depth'], ds['lidar-sd'], nrmse = False)
        res.loc[(site_name, 0), 'real_r'] = r
        res.loc[(site_name, 0), 'real_rmse'] = rmse
        res.loc[(site_name, 0), 'mean_snowdepth'] = ds['lidar-sd'].mean().data.ravel()[0]
        res.loc[(site_name, 0), 'fcf_mean'] = ds['fcf'].mean().data.ravel()[0]
        res.loc[(site_name, 0), 'dry_percentage'] = ds['wet_snow'].mean().data.ravel()[0]
        res.loc[(site_name, 0), 'valid-sd'] = (~ds['lidar-sd'].isnull()).sum().data.ravel()[0]


        rand_rmse_ds, rand_mae_ds, rand_r_ds = optimize(full_ds, A, B, C, im_date)

        a_best = rand_r_ds.max(['B', 'C']).idxmax('A')
        b_best = rand_r_ds.max(['C', 'A']).idxmax('B')
        c_best = rand_mae_ds.sel(A = a_best, B = b_best).idxmin('C')
        
        full_ds = s1_to_sd(full_ds, A = a_best, B = b_best, C = c_best)

        r, b, mae, rmse = get_stats(full_ds.sel(time = im_date, method = 'nearest')['snow_depth'], ds['lidar-sd'], nrmse = False)
        res.loc[(site_name, 0), 'optimized_r'] = r
        res.loc[(site_name, 0), 'optimized_rmse'] = rmse

    for cond in ['invert', 'fcf_s1_random', 's1_random']:
        for round_i in range(50):
    
            random_ds = full_ds.copy(deep = True)
            random_ds= random_ds.rio.write_crs('EPSG:4326')
            
            if cond == 'invert':
                random_ds = invert_timerseries(random_ds)[['s1', 'ims', 'fcf']]

            if cond == 'fcf_s1_random' or cond == 's1_random':
                random_ds = reorder_timeseries(random_ds, ims = False)[['s1', 'ims', 'fcf']]
            # elif cond == 's1_random':
            #     random_ds = reorder_timeseries(random_ds, ims = False)[['s1', 'ims', 'fcf']]
            if cond == 'fcf_s1_random':
                # random_ds['fcf'] = normalize_da(create_random_field(convert_to_utm(random_ds['fcf'])).rio.reproject_match(random_ds['fcf']))
                random_ds['fcf'] = normalize_da(create_random_field(convert_to_utm(random_ds['fcf']).where(convert_to_utm(random_ds['fcf']) < 10000)).rio.reproject_match(random_ds['fcf']))

            random_ds['lidar-sd'] = ds['lidar-sd']

            rand_rmse_ds, rand_mae_ds, rand_r_ds = optimize(random_ds, A, B, C, im_date)

            a_best = rand_r_ds.max(['B', 'C']).idxmax('A')
            b_best = rand_r_ds.max(['C', 'A']).idxmax('B')
            c_best = rand_mae_ds.sel(A = a_best, B = b_best).idxmin('C')
            
            r_ds_full = s1_to_sd(random_ds, A = a_best, B = b_best, C = c_best)
            
            r_sd = r_ds_full.where(r_ds_full['ims'] == 4)['snow_depth'].dropna(dim = 'time', how = 'all').sel(time = im_date, method = 'nearest')

            rand_r, rand_b, rand_mae, rand_rmse = get_stats(r_sd, ds['lidar-sd'], nrmse = False)
            fischer_z = fischerz(ds['lidar-sd'].data.ravel(), ds['snow_depth'].data.ravel(), r_sd.data.ravel())

            res.loc[(cond, round_i), 'random_r'] = rand_r
            res.loc[(cond, round_i), 'random_rmse'] = rand_rmse
            res.loc[(cond, round_i), 'fischerz'] = fischer_z 

            res.to_csv(outfp)

from multiprocessing import Pool
from functools import partial

print('starting')
in_dir = Path('~/scratch/spicy/SnowEx-Data/').expanduser().resolve()
data_dir = Path('~/scratch/spicy/SnowEx-Data/').expanduser().resolve()
out_dir = Path('/bsuhome/zacharykeskinen/spicy-analysis/results/synthetic_compare')

# Create parameter space
A = np.round(np.arange(1, 3.1, 0.5), 2)
B = np.round(np.arange(0, 2.01, 0.25), 2)
C = np.round(np.arange(0.25, 1.001, 0.25), 2)


A = np.round(np.arange(1, 3.1, 0.5), 2)
B = np.round(np.arange(0, 2.01, 0.25), 2)
C = np.round(np.arange(0.25, 1.001, 0.25), 2)

# out_dir.joinpath('sites').mkdir(exist_ok = True)

pool = Pool()

pool.map(partial(make_random_results, out_dir.joinpath('inverted'), A, B, C), list(in_dir.glob('*.nc')))


from asf_snow.utils.analyze_sentinel1_snow import *

def draw_s1_snotel_plot(file):

    lst = Path(file).stem.split("_")
    loc=lst[1].capitalize()
    res = int(lst[4])
    timeint = int(lst[5][:-1])
    timethresh = int(lst[6][:-1])
    base = lst[7]
    method = lst[8]
    precip = lst[9]

    outfile = file.replace(".geojson", ".png")

    gdf = gpd.read_file(file)

    # draw diff, error, and count plots
    # draw_s1_snotel_diff_error_count_plot(gdf, res, timeint, timethresh, base, method, outfile, precip, loc)

    draw_s1_sd_cr_snotel_plot(gdf, res, timeint, timethresh, base, method, outfile, precip, loc)

    # draw histogram
    hist_file = file.replace(".geojson", "_histogram.png")
    draw_histgram(gdf, res, timeint, timethresh, base, method, hist_file, precip, loc)

    return

if __name__ == '__main__':

    # file = "/home/jiangzhu/data/crrel/SNOTEL/creamersfield/s1_2/sentinel1_creamers_20220801_20230731n_400_12h_12h_s1_average_y_creamers_field.shp"

    parser = ArgumentParser()
    parser.add_argument('--file',type=str, required=True)
    args = parser.parse_args()
    draw_s1_snotel_plot(args.file)


import numpy

import logging
log = logging.getLogger(__name__)

def to01(array):

    log.debug("Making {array} to 0-1 range.")

    a = array.min()
    # ignore the Runtime Warning

    with numpy.errstate(divide='ignore'):
        b = 1. /(array.max() - array.min())

    if not(numpy.isfinite(b)):
        b = 0
        
    return numpy.vectorize(lambda x: b * (x - a))(array)


from pathlib import Path
from os.path import join
import shapely
import json
import glob
import logging
import os
import sys
from typing import Optional
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from asf_snow.download.sentinel1 import s1_img_search, hyp3_pipeline, download_hyp3, combine_s1_images, files2dataarray

log = logging.getLogger(__name__)

def main():
    parser = ArgumentParser()
    parser.add_argument('--username', type=str, default=None)
    parser.add_argument('--password', type=str, default=None)
    parser.add_argument('--area', nargs="*", type=float, default=[-113.2, 43, -113, 43.4])
    parser.add_argument('--daterange', nargs="*", type=str, default=['2008-08-01', '2021-04-01'])
    parser.add_argument('--workdir', type=str, default='.')
    parser.add_argument('--jobname', type=str, default='testing_job')
    parser.add_argument('--existjobname', type=str, default='testing_job')
    parser.add_argument('--outnc', default='test.nc')
    parser.add_argument('--debug', type=bool, default=False)

    args = parser.parse_args()

    # change to your minimum longitude, min lat, max long, max lat
    area = args.area

    area = shapely.geometry.box(area[0], area[1], area[2], area[3])

    workdir = args.workdir

    Path(workdir).mkdir(parents=True, exist_ok=True)

    # workdir = Path(workdir).expanduser()
    # work_dir =str(workdir)
    outdir = join(workdir, 'tmp')

    # this will be where your results are saved
    outnc = args.outnc
    out_nc = Path(outnc).expanduser()

    # this will generate a tuple of dates from the previous August 1st to this date
    # dates = get_input_dates(dates)  # run on all s1 images from (2020-08-01, 2021-04-01) in this example
    dates = args.daterange

    dates = tuple(dates)

    job_name = args.jobname

    existing_job_name = args.existjobname

    # spicy_ds = sentinel1_snow(area=args.area, dates=args.daterange, workdir=args.workdir,
    #                            jobname=args.jobname, existjobname=args.existjobname, outnc=args.outnc)

    # spicy_ds = retrieve_snow_depth(area=area1, dates=dates, work_dir=Path(workdir).expanduser(),
    #                               job_name=jobname,
    #                               existing_job_name=existjobname,
    #                               outfp=out_nc,
    #                               debug=False,
    #                               )

    # get asf_search search results
    search_results = s1_img_search(area, dates)
    log.info(f'Found {len(search_results)} results')

    num = len(search_results)

    if len(glob.glob(f'{outdir}/*.zip')) == num or len(glob.glob(f'{outdir}/S1*_VV.tif')) == num:
        if len(glob.glob(f'{outdir}/S1*_VV.tif')) == num:
            filelist = glob.glob(f'{outdir}/S1*_VV.tif')
        else:
            filelist = glob.glob(f'{outdir}/*.zip')
        imgs = files2dataarray(filelist, area=area, outdir=outdir, clean=False)
    else:

        assert len(search_results) > 3, f"Need at least 4 images to run. Found {len(search_results)} using area: {area} and dates: {dates}."

        # download s1 images into dataset ['s1'] variable name
        jobs = hyp3_pipeline(search_results, job_name=job_name, existing_job_name=existing_job_name)

        imgs = download_hyp3(jobs, area, outdir=join(workdir, 'tmp'), clean=False)


if __name__ == '__main__':
    main()



from asf_snow.utils.combine_multiple_statistic_results import *

file="/media/jiangzhu/Elements/crrel/sentinel1_snowEx21_ts_snow_pits_v1/boise/boise_202011_202104_500_24h_12h_s1_average_y.txt"

draw_snow_difference_histgram(file)





import geopandas as gpd
import xarray as xr
import xvec

from geodatasets import get_path

counties = gpd.read_file(get_path("geoda.natregimes"))

pop1960 = xr.DataArray(counties.PO60, coords=[counties.geometry], dims=["county"])


pop1960 = pop1960.xvec.set_geom_indexes("county", crs=counties.crs)








print("completed ...")






"""
Main user function to retrieve snow depth with snow depth and wet snow flag
"""
import os
from os.path import join
from pathlib import Path
import pandas as pd
import xarray as xr
import shapely.geometry
from typing import Tuple, Union, List
import logging

# Add main repo to path
import sys
from os.path import expanduser
sys.path.append(expanduser('../'))

# import functions for downloading
from spicy_snow.download.sentinel1 import s1_img_search, hyp3_pipeline, download_hyp3, combine_s1_images
from spicy_snow.download.forest_cover import download_fcf
from spicy_snow.download.snow_cover import download_snow_cover

# import functions for pre-processing
from spicy_snow.processing.s1_preprocessing import merge_partial_s1_images, s1_orbit_averaging,\
s1_clip_outliers, subset_s1_images, ims_water_mask, s1_incidence_angle_masking, merge_s1_subsets, \
add_confidence_angle

# import the functions for snow_index calculation
from spicy_snow.processing.snow_index import calc_delta_VV, calc_delta_cross_ratio, \
    calc_delta_gamma, clip_delta_gamma_outlier, calc_snow_index, calc_snow_index_to_snow_depth

# import the functions for wet snow flag
from spicy_snow.processing.wet_snow import id_newly_frozen_snow, id_newly_wet_snow, \
    id_wet_negative_si, flag_wet_snow

# setup root logger
from spicy_snow.utils.spicy_logging import setup_logging

def retrieve_snow_depth(area: shapely.geometry.Polygon, 
                        dates: Tuple[str, str], 
                        work_dir: str = './',
                        job_name: str = 'spicy-snow-run',
                        existing_job_name: Union[bool, str] = False,
                        debug: bool = False,
                        ims_masking: bool = True,
                        wet_snow_thresh: float = -2,
                        freezing_snow_thresh: float = 1,
                        wet_SI_thresh: float = 0,
                        outfp: Union[str, Path, bool] = False,
                        params: List[float] = [2.5, 0.2, 0.55]) -> xr.Dataset:
    """
    Finds, downloads Sentinel-1, forest cover, water mask (not implemented), and 
    snow coverage. Then retrieves snow depth using Lievens et al. 2021 method.

    Args:
    area: Shapely geometry to use as bounding box of desired area to search within
    dates: Start and end date to search between
    work_dir: filepath to directory to work in. Will be created if not existing
    job_name: name for hyp3 job
    existing_job_name: name for preexisiting hyp3 job to download and avoid resubmitting
    debug: do you want to get verbose logging?
    ims_masking: do you want to mask pixels by IMS snow free imagery?
    wet_snow_thresh: what threshold in dB change to use for melting and re-freezing snow? Default: -2
    freezing_snow_thresh: what threshold in dB change to use for re-freezing snow id. Default: +2
    wet_SI_thresh: what threshold to use for negative snow index? Default: 0
    outfp: do you want to save netcdf? default is False and will just return dataset
    params: the A, B, C parameters to use in the model. Current defaults are optimized to north america

    Returns:
    datset: Xarray dataset with 'snow_depth' and 'wet_snow' variables for all Sentinel-1
    image acquistions in area and dates
    """

    ## argument checking
    assert isinstance(area, shapely.geometry.Polygon), f"Must provide shapely geometry for area. Got {type(area)}"

    assert isinstance(dates, list) or isinstance(dates, tuple)
    assert len(dates) == 2, f"Can only provide two dates to work between. Got {dates}"

    assert isinstance(work_dir, str) or isinstance(work_dir, Path)
    if isinstance(work_dir, Path):
        work_dir = str(work_dir)
    
    assert isinstance(debug, bool), f"Debug keyword must be boolean. Got {debug}"

    assert isinstance(params, list) or isinstance(params, tuple), f"param keyword must be list or tuple. Got {type(params)}"
    assert len(params) == 3, f"List of params must be 3 in order A, B, C. Got {params}"
    A, B, C = params

    if type(outfp) != bool:
        outfp = Path(outfp).expanduser().resolve()
        assert outfp.parent.exists(), f"Out filepath {outfp}'s directory does not exist"

    ## set up directories and logging

    os.makedirs(work_dir, exist_ok = True)

    setup_logging(log_dir = join(work_dir, 'logs'), debug = debug)
    log = logging.getLogger(__name__)

    if wet_snow_thresh >= 0:
        log.warning(f"Running with wet snow threshold of {wet_snow_thresh}. This value is positive but should be negative.")
    
    if freezing_snow_thresh <= 0:
        log.warning(f"Running with refreeze threshold of {freezing_snow_thresh}. This value is negative but should be positive.")
    
    ## Downloading Steps

    # get asf_search search results
    search_results = s1_img_search(area, dates)
    log.info(f'Found {len(search_results)} results')

    assert len(search_results) > 3, f"Need at least 4 images to run. Found {len(search_results)} \
    using area: {area} and dates: {dates}."

    # download s1 images into dataset ['s1'] variable name
    jobs = hyp3_pipeline(search_results, job_name = job_name, existing_job_name = existing_job_name)
    imgs = download_hyp3(jobs, area, outdir = join(work_dir, 'tmp'), clean = False)
    ds = combine_s1_images(imgs)

    # merge partial images together
    ds = merge_partial_s1_images(ds)

    # download IMS snow cover and add to dataset ['ims'] keyword
    ds = download_snow_cover(ds, tmp_dir = join(work_dir, 'tmp'), clean = False)

    # download fcf and add to dataset ['fcf'] keyword
    ds = download_fcf(ds, join(work_dir, 'tmp', 'fcf.tif'))

    ## Preprocessing Steps
    log.info("Preprocessing Sentinel-1 images")

    #TODO add water mask
    # ds = ims_water_mask(ds)

    # mask out outliers in incidence angle
    ds = s1_incidence_angle_masking(ds)
    
    # subset dataset by flight_dir and platform
    dict_ds = subset_s1_images(ds)

    for subset_name, subset_ds in dict_ds.items():
        # average each orbit to overall mean
        dict_ds[subset_name] = s1_orbit_averaging(subset_ds)
        # clip outlier values of backscatter to overall mean
        dict_ds[subset_name] = s1_clip_outliers(subset_ds)
    
    # recombine subsets
    ds = merge_s1_subsets(dict_ds)

    # calculate confidence interval
    ds = add_confidence_angle(ds)

    ## Snow Index Steps
    log.info("Calculating snow index")
    # calculate delta CR and delta VV
    ds = calc_delta_cross_ratio(ds, A = A)
    ds = calc_delta_VV(ds)

    # calculate delta gamma with delta CR and delta VV with FCF
    ds = calc_delta_gamma(ds, B = B)

    # clip outliers of delta gamma
    ds = clip_delta_gamma_outlier(ds)

    # calculate snow_index from delta_gamma
    ds = calc_snow_index(ds, ims_masking = ims_masking)

    # convert snow index to snow depth
    ds = calc_snow_index_to_snow_depth(ds, C = C)

    ## Wet Snow Flags
    log.info("Flag wet snow")
    # find newly wet snow
    ds = id_newly_wet_snow(ds, wet_thresh = wet_snow_thresh)
    ds = id_wet_negative_si(ds, wet_SI_thresh = wet_SI_thresh)

    # find newly frozen snow
    ds = id_newly_frozen_snow(ds, freeze_thresh = freezing_snow_thresh)

    # make wet_snow flag
    ds = flag_wet_snow(ds)

    ds.attrs['param_A'] = A
    ds.attrs['param_B'] = B
    ds.attrs['param_C'] = C

    ds.attrs['job_name'] = job_name

    ds.attrs['bounds'] = area.bounds

    if outfp:
        outfp = str(outfp)
        
        ds.to_netcdf(outfp)

    return ds

def retrieval_from_parameters(dataset: xr.Dataset, 
                              A: float, 
                              B: float, 
                              C: float, 
                              wet_SI_thresh: float = 0, 
                              freezing_snow_thresh: float = 2,
                              wet_snow_thresh: float = -2):
    """
    Retrieve snow depth with varied parameter set from an already pre-processed
    dataset.

    Args:
    dataset: Already preprocessed dataset with s1, fcf, ims, deltaVV, merged images, 
    and masking applied.
    A: A parameter
    B: B parameter
    C: C parameter

    Returns:
    dataset: xarray dataset with snow_depth variable calculated from parameters
    """

    # dataset = dataset[['s1','deltaVV','ims','fcf','lidar-sd']]

    # load datast to index
    dataset = dataset.load()

    # calculate delta CR and delta VV
    dataset = calc_delta_cross_ratio(dataset, A = A)

    # calculate delta gamma with delta CR and delta VV with FCF
    dataset = calc_delta_gamma(dataset, B = B)

    # clip outliers of delta gamma
    dataset = clip_delta_gamma_outlier(dataset)

    # calculate snow_index from delta_gamma
    dataset = calc_snow_index(dataset)

    # convert snow index to snow depth
    dataset = calc_snow_index_to_snow_depth(dataset, C = C)

    # find newly wet snow
    dataset = id_newly_wet_snow(dataset, wet_thresh = wet_snow_thresh)
    dataset = id_wet_negative_si(dataset, wet_SI_thresh = wet_SI_thresh)

    # find newly frozen snow
    dataset = id_newly_frozen_snow(dataset, freeze_thresh =  freezing_snow_thresh)

    # make wet_snow flag
    dataset = flag_wet_snow(dataset)

    return dataset



# FOR GENERATING THE PARAMETER DATASETS! #

import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt
from pathlib import Path
from multiprocessing import Pool
from functools import partial
from datetime import datetime

from spicy_snow.processing.snow_index import calc_delta_cross_ratio, calc_delta_gamma, \
    clip_delta_gamma_outlier, calc_snow_index, calc_snow_index_to_snow_depth
from spicy_snow.processing.wet_snow import id_newly_wet_snow, id_wet_negative_si, \
    id_newly_frozen_snow, flag_wet_snow

def make_param_files(data_directory, idx, closest_ts, B, C, dataset, a):
    ds = dataset

    ds = calc_delta_cross_ratio(ds, A = a)
    for b in B:
        ds = calc_delta_gamma(ds, B = b, inplace=False)
        ds = clip_delta_gamma_outlier(ds)
        ds = calc_snow_index(ds)
        ds = id_newly_wet_snow(ds)
        ds = id_wet_negative_si(ds)
        ds = id_newly_frozen_snow(ds)
        ds = flag_wet_snow(ds)
        for c in C:
            ds = calc_snow_index_to_snow_depth(ds, C = c)

            sub = ds.sel(time = closest_ts)
            spicy_sd = sub['snow_depth'].values.ravel()[idx]
            # print(f'Saving {a}_{b}_{c}.npy')
            np.save(data_directory.joinpath(f'{a}_{b}_{c}.npy'), spicy_sd)

# Create parameter space
A = np.round(np.arange(1, 3.1, 0.1), 2)
B = np.round(np.arange(0, 2.01, 0.1), 2)
C = np.round(np.arange(0, 1.001, 0.01), 2)

spicy_netcdfs = list(Path('/bsuhome/zacharykeskinen/scratch/spicy/SnowEx-Data/').glob('*.nc'))

param_dir = Path('~/scratch/spicy/param_npys').expanduser()

param_dir.mkdir(exist_ok = True)

for spicy_nc in spicy_netcdfs:
    # get dataset
    ds_name = spicy_nc.name.split('stacks/')[-1].split('.')[0]
    print(datetime.now(), f' -- starting {ds_name}')
    ds_ = xr.open_dataset(spicy_nc).load()
    dataset = ds_[['s1','deltaVV','ims','fcf', 'lidar-sd', 'lidar-dem', 'snow_depth']]

    # find closest timestep to lidar
    td = abs(pd.to_datetime(dataset.time) - pd.to_datetime(dataset.attrs['lidar-flight-time']))
    closest_ts = dataset.time[np.argmin(td)]

    if 'Frasier_2020-02-11' in spicy_nc.name:
        closest_ts = '2020-02-16T13:09:43.000000000'

    ds_dir = param_dir.joinpath(ds_name)
    
    ds_dir.mkdir(exist_ok = True)

    trees = dataset['fcf'].values.ravel()
    elev = dataset['lidar-dem'].values.ravel()
    lidar = dataset['lidar-sd'].values.ravel()
    spicy_sd = dataset['snow_depth'].sel(time = closest_ts).values.ravel()
    wet = dataset['wet_snow'].sel(time = closest_ts).values.ravel()  # haven't tested wet saving so this may break in future

    idx = (~np.isnan(trees)) & (~np.isnan(elev)) & (~np.isnan(lidar)) & (~np.isnan(spicy_sd)) & (~np.isnan(wet)) # haven't tested wet saving so this may break in future
    # save auxillary files
    np.save(ds_dir.joinpath(f'lidar.npy'), lidar[idx])
    np.save(ds_dir.joinpath(f'trees.npy'), trees[idx])
    np.save(ds_dir.joinpath(f'elev.npy'), elev[idx])
    np.save(ds_dir.joinpath(f'wet.npy'), wet[idx])  # haven't tested wet saving so this may break in future


    # Brute-force processing loop
    pool = Pool()
        
    pool.map(partial(make_param_files, ds_dir, idx, closest_ts, B, C, dataset), A)


"""
Helper function to get bounding box from user
"""

import shapely.geometry
import xarray as xr
import rioxarray as rxa

from typing import Tuple, List

def get_input_area(coords: Tuple[List[str], List[float], bool] = False,
                   img: Tuple[xr.Dataset, xr.DataArray, bool] = False):
    """
    Helper function to return coordinates in bounding box format for retrieval
    Can only provide 1 argument type to this function.
    
    Args:
    coords: list of lat/long coordinates in format [xmin, ymin, xmax, ymax]
    img: xarray dataset or dataArray to extract bounding box from

    Returns:
    area: shapely.geometry.Polygon of bounding box
    """
    args = locals().values()

    assert any(args), "Must provide at least one argument"

    assert len([a for a in args if a != False]) == 1, "Must provide only argument."

    if coords:

        assert len(coords) == 4, f"Provide the 4 coordinates in [xmin, ymin, xmax, ymax] \
format. Provided list {coords} doesn't have 4 elements. Has {len(coords)}."

        if type(coords[0]) == str:
            coords = [float(c) for c in coords]
        
        xmin, ymin, xmax, ymax = coords

        assert xmin < xmax, f"xmin ({xmin}) is greater than xmax ({xmax})"
        assert ymin < ymax, f"ymin ({ymin}) is greater than ymax ({ymax})"

        assert ymin > -90, "Provide in lat/long"
        assert ymax < 90, "Provide in lat/long"

        assert xmin > -180, "Provide in lat/long"
        assert xmax < 180, "Provide in lat/long"

        area = shapely.geometry.box(xmin, ymin, xmax, ymax)

    if img:

        if img.rio.crs != 4326:
            img = img.rio.reproject("EPSG:4326")
        
        area = shapely.geometry.box(*img.rio.bounds())

    return area


from asf_snow.sentinel1_snow import *

from asf_snow.utils.analyze_sentinel1_snow import bbox_via_point

from asf_snow.utils.compare_sentinel1_snow_depth_to_snotel import *

from asf_snow.utils.draw_s1_snotel_plot import *



# 1, download SNOTEL manually

# 2, produce S1 data
# python sentinel1_snow.py --username cirrusasf --password xxxxxxxxx --area -147.75 64.85 -147.72 64.88 --daterange 2022-08-01 2023-07-31 --workdir /home/jiangzhu/data/crrel/SNOTEL/creamersfield/s1 --jobname jz_creamers_20220801_20230731_2 --existjobname jz_creamers_20220801_20230731_2 --outnc /home/jiangzhu/data/crrel/SNOTEL/creamersfield/s1/sentinel1_creamers_20220801_20230731_2.nc

# 3. product analysis results

# 4. draw plots




def main():
    parser = ArgumentParser()
    parser.add_argument('--snotelfile',type=str, required=True)
    parser.add_argument('--lon', type=float,required=True)
    parser.add_argument('--lat', type=float,required=True)
    parser.add_argument('--snotelarea', type=float,required=True)
    parser.add_argument('--username', type=str, default=None)
    parser.add_argument('--password', type=str, default=None)
    parser.add_argument('--daterange', nargs="*", type=str, default=['2008-08-01', '2021-04-01'])
    parser.add_argument('--workdir', type=str, default='./')
    parser.add_argument('--outnc', type=str, default='s1_data')
    parser.add_argument('--jobname', type=str, default='jz_s1')
    parser.add_argument('--existjobname', type=str, default='testing_job')
    parser.add_argument('--parameterabc', nargs="*", type=float, default=[2.5, 0.2, 0.55])
    parser.add_argument('--debug', type=bool, default=False)

    args = parser.parse_args()

    # download snotelfile mannually
    # Nenana
    # snotelfile = "/home/jiangzhu/data/crrel/SNOTEL/nenana/nenana_2081_20230801_20240731.csv"
    # lon=-148.91, lat=64.69, elevation=415 ft


    # lot,lat to bbox

    bbox = bbox_via_point(args.lon, args.lat, args.snotelarea)

    # produce S1 data
    # write args if args.workdir not exist

    environment_jsonfile = f'{args.workdir}/tmp/environment.json'

    if not Path(environment_jsonfile).is_file():
        environment = vars(args)
        environment['step'] = 'begin'
        environment['area'] = bbox
        os.environ['ENVIRONMENT'] = json.dumps(environment)

    else:
        environment = utils.read_environment(environment_jsonfile)
        os.environ['ENVIRONMENT'] = json.dumps(environment)

    # set environment for share
    if  args.username is not None and args.password is not None:
        os.environ['EARTHDATA_USERNAME'] = args.username
        os.environ['EARTHDATA_PASSWORD'] = args.password

    spicy_ds = sentinel1_snow(bbox, dates=args.daterange, workdir=args.workdir, jobname=args.jobname,
                              existjobname=args.jobname, outnc=args.outnc, parameterabc=args.parameterabc)

    # 3. analyze the s1 and snotel data

    resolutions = [50.0, 100.0, 200.0, 300.0, 400.0, 500.0]

    for res in resolutions:

        outfile = compare_sentinel1_snow_depth_to_snotel(args.outnc, 'snow_depth', args.snotelfile, args.lon, args.lat, res=res, timeint=24, timethresh=24, base='s1', method='average',
                                precip='n')
        # 4. draw plots
        draw_s1_snotel_plot(outfile)


if __name__ == '__main__':

    main()






"""
Functions to preprocess s1-images using methods of 2021 Lievens et al.

https://tc.copernicus.org/articles/16/159/2022/#section2
"""

from typing import Dict, List
import numpy as np
import pandas as pd
import xarray as xr
import rioxarray
from rioxarray.merge import merge_arrays
from itertools import product

import logging
log = logging.getLogger(__name__)

def s1_power_to_dB(dataset: xr.Dataset, inplace: bool = False):
    """
    Convert s1 images from amplitude to dB

    Args:
    dataset: Xarray Dataset of sentinel images in amplitude
    inplace: boolean flag to modify original Dataset or return a new Dataset

    Returns:
    dataset: Xarray dataset of sentinel image in dB
    """
    if not inplace:
        dataset = dataset.copy(deep=True)

    # check for dB
    if 's1_units' in dataset.attrs.keys():
        if dataset.attrs['s1_units'] == 'dB' and not inplace:
            log.info("Sentinel 1 units already in dB.")
            return dataset
        if dataset.attrs['s1_units'] == 'dB':
            return
    
    # mask all values 0 or negative
    dataset['s1'] = dataset['s1'].where(dataset['s1'] > 0)
    # convert all s1 images from amplitude to dB
    dataset['s1'].loc[dict(band = ['VV','VH'])] = 10 * np.log10(dataset['s1'].sel(band = ['VV','VH']))

    dataset.attrs['s1_units'] = 'dB'
    
    if not inplace:
        return dataset

def s1_dB_to_power(dataset: xr.Dataset, inplace: bool = False):
    """
    Convert s1 images from dB to amp

    Args:
    dataset: Xarray Dataset of sentinel images in dB
    inplace: boolean flag to modify original Dataset or return a new Dataset

    Returns:
    dataset: Xarray Dataset of sentinel images in amplitude
    """
    if not inplace:
        dataset = dataset.copy(deep=True)
    
    # check for amp
    if 's1_units' in dataset.attrs.keys():
        if dataset.attrs['s1_units'] == 'amp' and not inplace:
            print("Sentinel 1 units already in dB.")
            return dataset
        if dataset.attrs['s1_units'] == 'amp':
            return
        
    # convert all s1 images from amplitude to dB
    dataset['s1'].loc[dict(band = ['VV','VH'])] = 10 ** (dataset['s1'].sel(band = ['VV','VH']) / 10)
    dataset.attrs['s1_units'] = 'amp'
    if not inplace:
        return dataset


def merge_partial_s1_images(dataset, inplace: bool = False) -> xr.Dataset:
    """
    Merges s1 images that have been split by hyp3 into a single image with the 
    first time stamp of that relative orbit pass as the time index.

    Args:
    dataset: Xarray Dataset with Sentinel 1 images that have been arbitrarily
    split by hyp3 into an arbitrary number of subswaths

    Return:
    dataset: Xarray Dataset with Sentinel 1 images combined into single images and
    only the first subswath's time step
    """
    if not inplace:
        dataset = dataset.copy(deep=True)
    
    # split out each relative orbit and make its own dataset
    for orbit_num, orbit_ds in dataset.groupby('relative_orbit'):

        # split out each absolute orbit and combine images
        for absolute_num, abs_ds in orbit_ds.groupby('absolute_orbit'):

            # if only 1 image in this absolute orbit go to the next
            if len(abs_ds.time) == 1:
                continue

            # combine all images into the first image time step
            combo_imgs = abs_ds.mean(dim = 'time')

            # set first image in dataset to this new combined image
            dataset.loc[{'time' : abs_ds.isel(time = 0).time}] = combo_imgs

            # drop images that have been combined
            dataset = dataset.drop_sel(time = abs_ds.isel(time = slice(1, len(abs_ds.time))).time)

    # can leave some outliers in the dataset along the edges so remove unreasonable values
    if 's1' in dataset.data_vars:
        dataset['s1'] = dataset['s1'].where(dataset['s1'] < 100)
        dataset['s1'] = dataset['s1'].where(dataset['s1'] > -1e30)

    if not inplace:
        return dataset

def subset_s1_images(dataset: xr.Dataset) -> Dict[str, xr.Dataset]:
    """
    Subset s1 dataset into 4 subsets: ascending 1A, descending 1A, ascending 1B,
    and descending 1B

    Args:
    dataset: Xarray Dataset of sentinel images

    Returns:
    {'1A-asc': dataset, ...}: dictionary of up to 4 s1 datasets for permutations of 
    ascending, descending, and 1A and 1B with keys {satellite}-{direction}.
    """

    # options for platform and direction
    platforms, directions = ['S1A','S1B'], ['descending', 'ascending']

    # make subset dictionary
    subset_ds = {}

    # loop through possible combinations
    for platform, direction in product(platforms,directions):

        # subset on platform and direction
        subset = dataset.sel(time = dataset.platform == platform)
        subset = subset.sel(time = subset.flight_dir == direction)

        # save subset to dictionary
        if len(subset['s1']) > 0:
            subset_ds[f'{platform}-{direction}'] = subset
            log.debug(f"{platform}-{direction}: length = {len(subset_ds[f'{platform}-{direction}'])}")
    
    return subset_ds

def s1_orbit_averaging(dataset: xr.Dataset, inplace: bool = False) -> xr.Dataset:
    """
    Normalize s1 images by rescaling each image so its orbit's mean matches the
    overall time series mean. To allow for different orbits to be compared

    Args:
    dataset: Xarray Dataset of sentinel images to normalize by orbit 
    inplace: boolean flag to modify original Dataset or return a new Dataset

    Returns:
    dataset: Xarray Dataset of sentinel images with all s1 images normalized to total mean
    """

    # check inplace flag
    if not inplace:
        dataset = dataset.copy(deep=True)
    
    # check for dB
    if 's1_units' in dataset.attrs.keys():
        assert dataset.attrs['s1_units'] == 'dB', "Sentinel 1 units must be dB not amplitude."

    # get all unique relative orbits
    orbits = np.unique(dataset['relative_orbit'].values)

    # loop through bands
    for band in ['VV', 'VH']:
        
        # calculate the overall (all orbits) mean
        overall_mean  = dataset['s1'].mean(dim = ['x','y','time']).sel(band = band)
        log.debug(f"dataset's mean: {overall_mean}")

        for orbit in orbits:
            
            # calculate each orbit's mean value
            orbit_mean = dataset['s1'].sel(time = dataset.relative_orbit == orbit, band = band).mean(dim = ['x','y','time'])
            log.debug(f"Orbit's {orbit} pre-mean: {overall_mean}")

            # rescale each image by the mean correction (orbit mean -> overall mean)
            dataset['s1'].loc[dict(time = dataset.relative_orbit == orbit, band = band)] = \
                dataset['s1'].loc[dict(time = dataset.relative_orbit == orbit, band = band)] - (orbit_mean - overall_mean)

    if not inplace:
        return dataset

def s1_clip_outliers(dataset: xr.Dataset, inplace: bool = False) -> xr.Dataset:
    """
    Remove s1 image outliers by masking pixels 3 dB above 90th percentile or
    3 dB before the 10th percentile. (-35 -> 15 dB for VV) and (-40 -> 10 for VH
    in HL's code - line # 291)

    Args:
    dataset: Xarray Dataset of sentinel images to clip outliers
    inplace: boolean flag to modify original Dataset or return a new Dataset

    Returns:
    dataset: Xarray Dataset of sentinel images with masked outliers
    """
    # Check inplace flag
    if not inplace:
        dataset = dataset.copy(deep=True)

    # check for dB
    if 's1_units' in dataset.attrs.keys():
        assert dataset.attrs['s1_units'] == 'dB', "Sentinel 1 units must be dB not amplitude."

    # Calculate time series 10th and 90th percentile 
    # Threshold vals 3 dB above/below percentiles
    for band in ['VV','VH']:
        data = dataset['s1'].sel(band=band)

        thresh_lo, thresh_hi = data.quantile([0.1, 0.9], skipna = True)
        thresh_lo -= 3
        thresh_hi += 3
        # Mask using percentile thresholds
        data_masked = data.where((data > thresh_lo) & (data < thresh_hi))


        log.debug(f'Clipping band: {band}')
        log.debug(f'Thresh min: {thresh_lo.values}. Thresh max: {thresh_hi.values}')
        pre_min, pre_max = data.min().values, data.max().values
        log.debug(f'Data min: {pre_min}. Data max: {pre_max}')
        min, max = data_masked.min().values, data_masked.max().values
        log.debug(f'Masked data min: {min}. Data max: {max}')

        dataset['s1'].loc[dict(band = band)] = data_masked

    if not inplace:
        return dataset

def ims_water_mask(dataset: xr.Dataset) -> xr.Dataset:
    """
    Mask s1 pixels over water (1) or sea ice (3)

    Args:
    dataset: Xarray Dataset of sentinel images to clip outliers

    Returns:
    dataset: Xarray Dataset of sentinel images with masked water and sea ice
    """

    # mask all pixels in dataset where ims == 1 or 3.

def s1_incidence_angle_masking(dataset: xr.Dataset, inplace: bool = False) -> xr.Dataset:
    """
    Remove s1 image outliers by masking pixels with incidence angles > 70 degrees
    
    Args:
    dataset: Xarray Dataset of sentinel images to mask

    Returns:
    dataset: Xarray Dataset of sentinel images with incidence angles > 70 degrees
    masked
    """
    
    # Check inplace flag
    if not inplace:
            dataset = dataset.copy(deep=True)

    # Mask pixels with incidence angle > 70 degrees
    dataset['s1'] = dataset['s1'].where(dataset['s1'].sel(band = 'inc') < np.deg2rad(70))

    if not inplace:
            return dataset

def merge_s1_subsets(dataset_dictionary: Dict[str, xr.Dataset]) -> xr.Dataset:
    """
    Remove s1 image outliers by masking pixels with incidence angles > 70 degrees

    Args:
    {'1A-asc': dataset, ...}: dictionary of up to 4 s1 datasets for permutations of 
    ascending, descending, and 1A and 1B with keys {satellite}-{direction}.

    Returns:
    dataset: Xarray Dataset of all preprocessed sentinel images
    """
    # merge subsets of orbit/satellite into one Dataset
    dataset = xr.merge(dataset_dictionary.values())

    return dataset

def add_confidence_angle(dataset: xr.Dataset, inplace: bool = False):
    """
    Function to add confidence angle to dataset.

    Confidence Angle = angle[ abs(dVH/dt / mean(dVH/dt)), abs(dVV/dt / mean(dVV/dt)) ]

    Args:
    dataset: Xarray Dataset of sentinel images to add confidence angle to
    inplace: boolean flag to modify original Dataset or return a new Dataset

    Returns:
    dataset: Xarray dataset of sentinel image with confidence interval in 
    """
    ds_amp = s1_dB_to_power(dataset).copy()
    ds_amp['deltaVH_amp'] = ds_amp['s1'].sel(band = 'VH').diff(1)
    ds_amp['deltaVV_amp'] = ds_amp['s1'].sel(band = 'VV').diff(1)

    ds_amp['deltaVH_norm'] = np.abs(ds_amp['deltaVH_amp'] / ds_amp['deltaVH_amp'].mean())
    ds_amp['deltaVV_norm'] = np.abs(ds_amp['deltaVV_amp'] / ds_amp['deltaVV_amp'].mean())

    ds_amp['confidence'] = (ds_amp['deltaVH_norm'].dims, np.angle(ds_amp['deltaVV_norm'].values + ds_amp['deltaVH_norm'].values * 1j))
    dataset['confidence'] = ds_amp['confidence'].mean('time')

    if not inplace:
        return dataset



"""
Functions to search and download Sentinel-1 images for specific geometries and dates
"""

import sys
import os
from os.path import basename, exists, expanduser, join
from pathlib import Path
from zipfile import ZipFile
import shutil
import asf_search as asf
import pandas as pd
import xarray as xr
import rioxarray as rxa
from rioxarray.merge import merge_arrays
import shapely.geometry
from datetime import date
from tqdm import tqdm
import hyp3_sdk as sdk
from hyp3_sdk.exceptions import AuthenticationError

from typing import Dict, Tuple, List, Union

import sys
from os.path import expanduser
sys.path.append(expanduser('~/Documents/spicy-snow'))
from asf_snow.utils.download import url_download
from asf_snow.utils.utils import *
from spicy_snow.processing.s1_preprocessing import s1_power_to_dB

import logging
log = logging.getLogger(__name__)


def s1_img_search(area: shapely.geometry.Polygon, dates: Tuple[str, str]) -> pd.DataFrame:
    """
    find dates and url of Sentinel-1 overpasses

    Args:
    area: Bounding box of desired area to search within
    dates: Start and end date to search between

    Returns:
    granules: Dataframe of Sentinel-1 granule names to download.
    """
    # Error Checking
    if len(dates) != 2:
        raise TypeError("Provide at start and end date in format (YYYY-MM-DD, YYYY_MM_DD)")
    if type(area) != shapely.geometry.Polygon:
        raise TypeError("Geometry must be a shapely.geometry.box type")
    if type(dates[0]) != str:
        raise TypeError("Provide at start and end date in format (YYYY-MM-DD, YYYY_MM_DD)")
    dates = [pd.to_datetime(d) for d in dates]
    if dates[1] < dates[0]:
        raise ValueError("End date is before start date")
    if dates[0].year < 2014 or dates[1].year < 2014:
        raise IndexError("Dates are prior to Sentinel-1 launch dates")
    if dates[0].date() > date.today() or dates[1].date() > date.today():
        raise IndexError("Dates are in the future.")
    if area.bounds[3] > 90 or area.bounds[1] < 0 or area.bounds[2] > 180\
        or area.bounds[0] < -180:
        raise IndexError("Coordinates must be between 0-90N and -180-180")

    # get results from asf_search in date range and geometry
    results = asf.geo_search(platform = [asf.PLATFORM.SENTINEL1], intersectsWith = area.wkt,\
        start = dates[0], end = dates[1], processingLevel = asf.PRODUCT_TYPE.GRD_HD)
    
    # check with 0 results.
    if len(results) == 0:
        raise ValueError("No search results found.")

    # create pandas dataframe from json result
    results = pd.json_normalize(results.geojson(), record_path = ['features'])

    return results


def hyp3_pipeline(search_results: pd.DataFrame, job_name, existing_job_name: Union[bool, str] = False) -> sdk.jobs.Batch:
    """
    Start and monitor Hyp3 pipeline for desired Sentinel-1 granules
    https://hyp3-docs.asf.alaska.edu/using/sdk_api/

    Args:
    search_results: Pandas Dataframe of asf_search search results.
    job_name: name to give hyp3 batch run
    existing_job_name: if you have an existing job that you want to find and reuse [default: False]

    Returns:
    rtc_jobs: Hyp3 batch object of completed jobs.
    """ 
    try:
        # .netrc
        hyp3 = sdk.HyP3()
    except AuthenticationError:
        # prompt for password
        hyp3 = sdk.HyP3(prompt = True)

    # if existing job name exists then don't submit and simply watch existing jobs.
    while existing_job_name:
        log.debug(f"existing name provided {existing_job_name}.")
        rtc_jobs = hyp3.find_jobs(name = existing_job_name)
        rtc_jobs = rtc_jobs.filter_jobs(succeeded = True, failed = False, \
            running = True, include_expired = False)
        log.debug(f"Found {len(rtc_jobs)} jobs under existing name. \
                  This is only succeeded and running jobs.")

        # if no jobs found go to original search with name.
        if len(rtc_jobs) == 0:
            break

        # if no running jobs then just return succeeded jobs
        if len(rtc_jobs.filter_jobs(succeeded = False, failed = False)) == 0:
            log.debug("No running jobs. Returning succeeded.")
            return rtc_jobs.filter_jobs(succeeded = True)

        # otherwise watch running jobs
        hyp3.watch(rtc_jobs)

        # refresh with new successes and failures
        rtc_jobs = hyp3.refresh(rtc_jobs)
        
        log.debug(f"Successful jobs returned after watching {len(rtc_jobs.filter_jobs(succeeded = True))}")
        # return successful jobs
        return rtc_jobs.filter_jobs(succeeded = True)
    
    # check if you have passed quota
    quota = hyp3.check_credits()
    if not quota or len(search_results) > quota:
        log.warn(f'More search results ({len(search_results)}) than quota ({quota}).')
        resp = None
        while resp not in ['Y', 'N']:
            resp = input('Continue anyways?')[:1].upper()
            if resp not in ['Y', 'N']:
                print('Enter Y or N.')
        
        if resp == 'N':
            sys.exit("Not enough jobs left in ASF Hyp3 quota.")


    # gather granules to submit to the hyp3 pipeline
    granules = search_results['properties.sceneName']

    # create a new hyp3 batch to hold submitted jobs
    rtc_jobs = sdk.Batch()
    for g in tqdm(granules, desc = 'Submitting s1 jobs'):
        # submit rtc jobs and ask for incidence angle map, in dBs, @ 30 m resolution
        # https://hyp3-docs.asf.alaska.edu/using/sdk_api/#hyp3_sdk.hyp3.HyP3.submit_rtc_job
        rtc_jobs += hyp3.submit_rtc_job(g, name = job_name, include_inc_map = True,\
            scale = 'power', dem_matching = False, resolution = 30)

    # warn user this may take a few hours for big jobs
    log.info(f'Watching {len(rtc_jobs)} jobs. This may take a while...')

    # have hyp3 watch and update progress bar every 60 seconds
    hyp3.watch(rtc_jobs)

    # refresh jobs list with successes and failures
    rtc_jobs = hyp3.refresh(rtc_jobs)

    # filter out failed jobs
    failed_jobs = rtc_jobs.filter_jobs(succeeded=False, running=False, failed=True)
    if len(failed_jobs) > 0:
        log.info(f'{len(failed_jobs)} jobs failed.')
    
    # return only successful jobs
    return rtc_jobs.filter_jobs(succeeded = True)


def download_hyp3(jobs: sdk.jobs.Batch, area: shapely.geometry.Polygon, outdir: str, clean = True) -> Dict[str, xr.DataArray]:
    """
    Download rtc Sentinel-1 images from Hyp3 pipeline.
    https://hyp3-docs.asf.alaska.edu/using/sdk_api/

    Args:
    jobs: hyp3 Batch object of completed jobs
    outdir: directory to save tif files.
    clean: clean up tiffs after creating DataArray [default: True]

    Returns:
    images: dictionary of granule names and DataArrays
    """
    log.debug(f"Downloading hyp3 jobs into {outdir}")
    # make data directory to store incoming tifs
    os.makedirs(outdir, exist_ok = True)

    # results dictionary to send to next step
    dataArrays = {}

    # grab first granule name for reprojecting matching
    first_granule = jobs[0].job_parameters['granules'][0]

    # loop through jobs
    for job in tqdm(jobs, desc = 'Downloading S1 images'):
        # capture url from job description
        u = job.files[0]['url']

        # capture granule (for metadata scraping)
        granule = job.job_parameters['granules'][0]

        # skip this loop if granule is repeated in job list
        if granule in dataArrays.keys():
            continue

        # create dictionary to hold cloud url from .zip url
        # this lets us download only VV, VH, inc without getting other data from zip
        urls = {}
        urls[f'{granule}_VV'] = u.replace('.zip', '_VV.tif')
        urls[f'{granule}_VH'] = u.replace('.zip', '_VH.tif')
        urls[f'{granule}_inc'] = u.replace('.zip', '_inc_map.tif')

        imgs = []

        for name, url in urls.items():
            # download url to a tif file
            if url_download(url, join(outdir, f'{name}.tif'), verbose=False):
                # open image in xarray
                img = rxa.open_rasterio(join(outdir, f'{name}.tif'), masked=True)

                # reproject to WGS84
                img = img.rio.reproject('EPSG:4326')

                # clip to user specified area
                img = img.rio.clip_box(*area.bounds)

                # pad to user specified area
                img = img.rio.pad_box(*area.bounds)

                # create band name
                band_name = name.replace(f'{granule}_', '')

                # add band to image
                img = img.assign_coords(band = [band_name])

                # add named band image to 3 image stack
                imgs.append(img)

        # concat VV, VH, and inc into one xarray DataArray
        da = xr.concat(imgs, dim = 'band')

        # coarsen to correct resolution (90 m)
        da = da.coarsen(x = 3, boundary = 'trim').mean().coarsen(y = 3, boundary = 'trim').mean()

        # we need to reproject each image to match the first image to make CRSs work
        if dataArrays:
            da = da.rio.reproject_match(dataArrays[first_granule])

        # add img to downloaded dataArrays list with granule as key
        dataArrays[granule] = da
    
    # remove temp directory of tiffs
    if clean:
        shutil.rmtree(outdir)

    return dataArrays



def files2dataarray(filelist: list, area: shapely.geometry.Polygon, outdir: str, clean=True) -> Dict[str, xr.DataArray]:
    """ Read rtc Sentinel-1 images and convert them into dict of [str, xr.DataArray].
    Args:
    filelist
    outdir: directory to save tif files.
    clean: clean up tiffs after creating DataArray [default: True]

    Returns:
    images: dictionary of granule names and DataArrays
    """
    log.debug(f"Downloading hyp3 jobs into {outdir}")
    # make data directory to store incoming tifs
    os.makedirs(outdir, exist_ok=True)

    # results dictionary to send to next step
    dataArrays = {}

    # check file type of the context in the filelist
    # grab first granule name for reprojecting matching
    if filelist[0].find(".zip") > 0:
        filetype = "zip"
        first_granule = Path(filelist[0]).name.split(".zip")[0]
    else:
        filetype = "tif"
        first_granule = Path(filelist[0]).name.split("_VV.tif")[0]


    # loop through jobs
    for job in tqdm(filelist, desc=f'read the S1 {filetype} files'):
        # capture url from job description
        # u = job.files[0]['url']

        # capture granule (for metadata scraping)
        # granule = job.job_parameters['granules'][0]
        if filetype == "zip":
            granule = Path(job).name.split(".zip")[0]
            # create dictionary to hold cloud url from .zip url
            # this lets us download only VV, VH, inc without getting other data from zip
            urls = {}
            urls[f'{granule}_VV'] = job.replace('.zip', '_VV.tif')
            urls[f'{granule}_VH'] = job.replace('.zip', '_VH.tif')
            urls[f'{granule}_inc'] = job.replace('.zip', '_inc_map.tif')
        else:
            granule = Path(job).name.split("_VV.tif")[0]
            urls = {}
            urls[f'{granule}_VV'] = job.replace('_VV.tif', '_VV.tif')
            urls[f'{granule}_VH'] = job.replace('_VV.tif', '_VH.tif')
            urls[f'{granule}_inc'] = job.replace('_VV.tif', '_inc.tif')

        # skip this loop if granule is repeated in job list
        if granule in dataArrays.keys():
            continue

        imgs = []

        for name, url in urls.items():
            # download url to a tif file
            # url_download(url, join(outdir, f'{name}.tif'), verbose=False)
            if filetype == "zip":
                file = extractfromzipfile(Path(job), Path(url))
            else:
                file = Path(url)

            try:
                # open image in xarray
                img = rxa.open_rasterio(file, masked=True)

                # reproject to WGS84
                img = img.rio.reproject('EPSG:4326')

                # clip to user specified area
                img = img.rio.clip_box(*area.bounds)

                # pad to user specified area
                img = img.rio.pad_box(*area.bounds)

                # create band name
                band_name = name.replace(f'{granule}_', '')

                # add band to image
                img = img.assign_coords(band=[band_name])

                # add named band image to 3 image stack
                imgs.append(img)

            except:
                continue

        # concat VV, VH, and inc into one xarray DataArray
        da = xr.concat(imgs, dim='band')

        # coarsen to correct resolution (90 m)
        da = da.coarsen(x=3, boundary='trim').mean().coarsen(y=3, boundary='trim').mean()

        # we need to reproject each image to match the first image to make CRSs work
        if dataArrays:
            da = da.rio.reproject_match(dataArrays[first_granule])

        # add img to downloaded dataArrays list with granule as key
        dataArrays[granule] = da

    # remove temp directory of tiffs
    if clean:
        shutil.rmtree(outdir)

    return dataArrays


def combine_s1_images(dataArrays: Dict[str, xr.DataArray]) -> xr.Dataset:
    """
    Combine list of 3-banded Sentinel 1 data Arrays into a single xarray
    Dataset with associated metadata bands and attributes.

    Args:
    dataArrays: dictionary of granule name and 3 band Sentinel-1 data Arrays

    Returns:
    dataset: xr dataset with time dimension added, metadata attributes, and 
    metadata coordinates (flight direction, orbit #, platform)
    """
    das = []

    for granule, da in tqdm(dataArrays.items(), desc = 'Combining Sentinel-1 dataArrays'):
        # get granule metadata
        granule_metadata = asf.product_search(f'{granule}-GRD_HD')[0]

        # set flight direction
        flight_dir = granule_metadata.properties['flightDirection'].lower()

        # set relative orbit 
        relative_orbit = granule_metadata.properties['pathNumber']

        # set absolute orbit
        absolute_orbit = granule_metadata.properties['orbit']

        # expand time dimension of DataArray from zero dimension (scalar) to 1d
        da = da.expand_dims(dim = {'time': 1})

        # add time as a indexable parameter
        da = da.assign_coords(time = [pd.to_datetime(granule.split('_')[4])])

        # add flight direction as indexable parameter
        da = da.assign_coords(flight_dir = ('time', [flight_dir]))

        # add platform as indexable parameter
        platform = granule[0:3]
        da = da.assign_coords(platform = ('time', [platform]))

        # add relative orbit as indexable parameter
        da = da.assign_coords(relative_orbit = ('time', [relative_orbit]))

        # add absolute orbit as indexable parameter
        da = da.assign_coords(absolute_orbit = ('time', [absolute_orbit]))
        
        # append multi-band image to das list to concat into time-series DataArray
        das.append(da)

    # take list of multi-band images with different time values and make time series
    s1_dataArray = xr.concat(das, dim = 'time')

    # make sentinel 1 dataset 
    s1_dataset = s1_dataArray.to_dataset(name = 's1', promote_attrs = True)

    # resolution set to 90m:
    # must do in linear space not logarithmic dBs
    assert s1_dataset['s1'].sel(band = ['VV', 'VH']).min() >= 0
    # s1_dataset = s1_dataset.coarsen(x = 3, boundary = 'trim').mean().coarsen(y = 3, boundary = 'trim').mean()
    s1_dataset.attrs['resolution'] = '90'

    # ensure time dimension is sorted
    s1_dataset = s1_dataset.sortby('time')
    
    # s1_units tag
    s1_dataset = s1_power_to_dB(s1_dataset)
    s1_dataset.attrs['s1_units'] = 'dB'

    return s1_dataset

# End of file



from spicy_snow.retrieval import retrieve_snow_depth

from pathlib import Path

from spicy_snow.IO.user_dates import get_input_dates

from shapely.geometry import box
area = box(-115.574500,43.473139,-113.970496,44.418369)

dates = get_input_dates('2021-05-01')

data_dir = Path('/bsuhome/zacharykeskinen/scratch/idaho_full/')

spicy_ds = retrieve_snow_depth(area = area, dates = dates, work_dir = data_dir, \
                                           job_name = f'idaho', existing_job_name = f'idaho', outfp = data_dir.joinpath('idaho.nc'))


from pathlib import Path
import sys
import os
import glob
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import numpy as np
import numpy.ma as ma
import xarray
import rioxarray as rio
from rasterio.enums import Resampling
from osgeo import gdal
from asf_snow.utils.utils import *
import xml.etree.ElementTree as ET
from datetime import datetime, timedelta

import mpl_scatter_density # adds projection='scatter_density'
from matplotlib.colors import LinearSegmentedColormap

from asf_snow.utils.analyze_sentinel1_snow import *
from asf_snow.utils.analyze_snex import *

import pickle

# compare the snow depth derived with sentinel1 to NASA SnowEx experiment Lidar snow depth data
# example arguments
# ncfile = "/home/jiangzhu/data/crrel/sentinel1_snowEx21_ts_snow_pits_v1/sentinel1.nc"
# lidarfile = "/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/SNEX20_QSI_SD_0.5M_USIDBS_20200218_20200219.tif"
# varname = "snow_depth"
# res = 100  # zonal analysis square size

def get_geotransform(data:xarray):
    gt_str = data.spatial_ref.GeoTransform
    gt = [float(item) for item in gt_str.split(' ')]
    return gt

def get_bounds(gt, xsize, ysize):
    xn = gt[0] + gt[1]*xsize
    yn = gt[3] + gt[5]*ysize
    return [gt[0], yn, xn, gt[3]]


def resample(src: xarray.DataArray,  ref: xarray.DataArray):
    # use gdawrap to resample

    dst_xsize = ref.shape[2]

    dst_ysize = ref.shape[1]

    dst_bound = ref.rio.bounds()

    #kwargs = {"width": dst_xsize, "height": dst_ysize, "outputBounds": dst_bound, "format": "GTiff"}

    src.rio.to_raster('/tmp/src.tif')

    #ds = gdal.Warp('/tmp/sample.tif', '/tmp/src.tif', **kwargs)

    #dst = rio.open_rasterio("/tmp/sample.tif")
    if os.path.exists('/tmp/sample.tif'):
        os.system('/bin/rm /tmp/sample.tif')
    # cmd = f'gdalwarp -te {dst_bound[0]}  {dst_bound[1]} {dst_bound[2]} {dst_bound[3]} -ts {dst_xsize} {dst_ysize} /tmp/src.tif /tmp/sample.tif'

    cmd = f'gdalwarp -r average -te {dst_bound[0]}  {dst_bound[1]} {dst_bound[2]} {dst_bound[3]} -ts {dst_xsize} {dst_ysize} /tmp/src.tif /tmp/sample.tif'
    # gdalwarp -te 637308.612295736 4894974.424383238 652555.8674305859 4911581.307364508 -ts 157 171 lidar.tif lidar_sample_te_ts.ti

    os.system(cmd)

    dst = rio.open_rasterio('/tmp/sample.tif')

    return dst

def calc_rmse(x, y):
    xm = ma.masked_invalid(x)
    ym = ma.masked_invalid(y)
    return np.sqrt(((xm - ym) ** 2).mean())


def calc_corr(y,x):
    """
    calculate the r and mbe
    :param y: ground true
    :param x: measurement
    :return: r, mbe
    """
    x = ma.masked_invalid(x)
    y = ma.masked_invalid(y)
    # if x and y are 2D, flatten them
    if x.ndim == 2:
        x = x.flatten()
        y = y.flatten()
    # get rid of nan data
    invalid_msk = x.mask | y.mask

    x = x[~invalid_msk]
    y = y[~invalid_msk]

    # calculate the correlation coefficent
    r = np.corrcoef(y,x)

    # calculate the mean bias error
    mbe = np.mean(y - x)
    return r[0,1], mbe

def str2datetime64(dtstr, format="%Y-%m-%d %H:%M:%S.%f"):
    dt = datetime.strptime(dtstr, format)
    dt64 = np.datetime64(dt)
    return dt64


def get_datetime_range(xml):
    tree = ET.parse(xml)

    root = tree.getroot()

    metadata  = root.findall('./GranuleURMetaData')

    rangedatetime = metadata[0].findall('./RangeDateTime')

    ed_date = rangedatetime[0].find('RangeEndingDate').text
    ed_time = rangedatetime[0].find('RangeEndingTime').text
    st_date = rangedatetime[0].find('RangeBeginningDate').text
    st_time = rangedatetime[0].find('RangeBeginningTime').text

    ed_str = f'{ed_date} {ed_time}'

    ed_dt64 = str2datetime64(ed_str)

    st_str = f'{st_date} {st_time}'

    st_dt64 = str2datetime64(st_str)

    return st_str, ed_str

def str2datetime(str, format="%Y-%m-%d %H:%M:%S.%f"):
    return datetime.strptime(str, format)


def scatter_plot(x, y, outfile):
    x = ma.masked_invalid(x)
    y = ma.masked_invalid(y)
    # if x and y are 2D, flatten them
    if x.ndim == 2:
        x = x.flatten()
        y = y.flatten()
    # get rid of nan data
    invalid_msk = x.mask | y.mask

    x = x[~invalid_msk]
    y = y[~invalid_msk]

    ch = np.bitwise_and(x >= 0, y >= 0)
    x = x[ch]
    y = y[ch]

    slope, intercept = np.polyfit(x, y, deg=1)
    ye = slope*x + intercept
    slope1 = round(slope, 2)
    intercept1 = round(intercept, 2)

    rmse = round(np.sqrt(((x - y) ** 2).mean()), 2)

    r = np.corrcoef(y,x)
    coeff = r[0,1]
    coeff = round(coeff, 2)

    n = x.shape[0]

    fig, ax = plt.subplots()
    fig.set_figheight(8)
    fig.set_figwidth(8)

    # histogram definition
    bins = [100, 100]  # number of bins

    # histogram the data
    hh, locx, locy = np.histogram2d(x, y, bins=bins)

    # Sort the points by density, so that the densest points are plotted last
    z = np.array([hh[np.argmax(a <= locx[1:]), np.argmax(b <= locy[1:])] for a, b in zip(x, y)])
    idx = z.argsort()
    x2, y2, z2 = x[idx], y[idx], z[idx]

    fig.clf()
    sc = plt.scatter(x2, y2, c=z2, cmap='jet', marker='.')
    ax = sc._axes
    ax.set_xlabel('Lidar Snow Depth (m)')
    ax.set_ylabel('S1 Snow Depth (m)')
    ax.plot(x,ye)

    # ax.legend(loc='best')  # in order to display label, you have to use ax.legend
    ax.text(0.5, 0.95, f'RMSE {rmse}, R {coeff}, n {n}', color='black', transform = ax.transAxes)
    ax.set_xlim((0,7))
    ax.set_ylim((0,7))
    plt.colorbar(sc, label="Pixel Counts")
    ax.legend()
    plt.show()

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)


def compare_sentinel1_snow_depth_to_lidar(ncfile, varname, lidarfile, location, res, timeint, timethresh, base, method,
                                         precip):

    work_dir = str(Path(ncfile).parent)

    # sentinel1 snow depth data
    nc_var = get_var(ncfile, varname)

    # NASA snowEx Lidar snow depth data

    lidar = rio.open_rasterio(lidarfile)

    xml = lidarfile +'.xml'

    st_dt64, ed_dt64 = get_datetime_range(xml)

    lidar = lidar.assign_attrs({'start_datetime': st_dt64, 'end_datetime': ed_dt64 })

    # reproject both lidar_data and nc_var to UTM

    utm = lidar.rio.estimate_utm_crs()

    nc_var = nc_var.rio.reproject(utm)

    lidar = lidar.rio.reproject(utm)

    # clip nc_var
    nc_var_clipped = nc_var.rio.clip_box(*lidar.rio.bounds())

    nc_var_clipped.rio.to_raster(f'{work_dir}/nc_var_clipped.tif')

    # resample lidar
    lidar_sample = resample(lidar, nc_var_clipped)

    lidar_sample.rio.to_raster(f'{work_dir}/lidar_sample.tif')

    # pick the image from the nc_var_clipped with the time is in the range of the st_time and ed_time
    # Banner, idaho, time zoom -6 hours
    t_st = str2datetime(lidar_sample.start_datetime) - timedelta(hours=6)
    t_ed = str2datetime(lidar_sample.end_datetime) - timedelta(hours=6)

    nc_timerange = nc_var_clipped.sel(time=slice(t_st, t_ed))

    # clip both lidar_sample and nc_timerange with shrink lidar_sample_outline, get rid of edge pixels
    '''
    rasterfile='/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/lidar_sample.tif'
    outlinefile ='/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/lidar_sample_outline.shp'

    outlinefile = extract_outline_from_raster(rasterfile, outlinefile)

    shrinkfile = '/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/lidar_sample_shrink.shp'
    geoms = shrink_geometry(outlinefile, shrinkfile, buff= -200.0)

    nc_timerange = nc_timerange.rio.clip(geoms)

    nc_timerange.rio.to_raster('/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/nc_timerange_clipped.tif')

    lidar_sample = lidar_sample.rio.clip(geoms)

    lidar_sample.rio.to_raster('/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/lidar_sample_clipped.tif')
    
    '''
    # calculate the RMSE, R, and MBE

    rmse = calc_rmse(lidar_sample[0].data, nc_timerange[0].data)

    r, mbe = calc_corr(lidar_sample[0].data, nc_timerange[0].data)

    print(f'RMSE {rmse}, r {r}, mbe {mbe}')

    # draw spot plot and save it
    outfile = f'{work_dir}/{str(Path(ncfile).stem)}_{varname}_{str(Path(lidarfile).stem)}.png'

    f = open(f'{work_dir}/store.pckl', 'wb')
    pickle.dump([lidar_sample[0].data,nc_timerange[0].data], f)
    f.close()

    scatter_plot(lidar_sample[0].data, nc_timerange[0].data, outfile)


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('--ncfile', type=str,  required=True)
    parser.add_argument('--varname', type=str, default='snow_depth')
    parser.add_argument('--lidarfile', type=str,  required=True)
    parser.add_argument('--location', type=str, default='Boise')
    parser.add_argument('--res', type=float, default=100)
    parser.add_argument('--timeint', type=int, default=24)
    parser.add_argument('--timethresh', type=int, default=6)
    parser.add_argument('--base', choices=['snex', 's1'], default='s1')
    parser.add_argument('--method', choices=['nearest', 'average', 'close',
                                             'nearest', 'average', 'close'])
    parser.add_argument('--precip', choices=['y', 'n'], default='y')

    args = parser.parse_args()

    compare_sentinel1_snow_depth_to_lidar(args.ncfile, varname=args.varname, lidarfile=args.lidarfile,
                                         location=args.location, res=args.res, timeint=args.timeint,
                                         timethresh=args.timethresh, base=args.base, method=args.method,
                                         precip=args.precip)

    print('completed analysis...')



from asf_snow.utils.compare_sentinel1_snow_depth_to_snex_lidar import *


lidarfile = "/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/lidar_sample.tif"

ncfile = "/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/nc_timerange.tif"

lidar = rio.open_rasterio(lidarfile)

nc = rio.open_rasterio(ncfile)

scatter_plot(lidar[0].data, nc[0].data, "/home/jiangzhu/data/crrel/SnowEx20-21_QSI_Lidar_Snow_Depth_0d5m_UTM_Grid_V1/test/scatter.png")





from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
import xarray as xr
import rioxarray as rio
import rasterio
from folium import GeoJson
from shapely import geometry, to_geojson
import pandas as pd
import geopandas as gpd
import rasterstats as rstats
from affine import Affine
import numpy as np
import utm
from pyproj import CRS, Proj
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter, PercentFormatter

#from shapely.geometry import Polygon, LineString, Point

from datetime import datetime, timedelta

import json
import matplotlib

# matplotlib.use('Agg')

import pytz
from timezonefinder import TimezoneFinder

def get_utc_offset(lon, lat):
    tf = TimezoneFinder()
    tz_str = tf.timezone_at(lng=lon, lat=lat)
    timezone = pytz.timezone(tz_str)
    return timezone

def get_utc(datetime, tz):
    return datetime + tz.utcoffset(datetime)


def get_crs(lon: float, lat: float) -> CRS:
    y, x, zone, band = utm.from_latlon(lat, lon)
    if lat < 0.0:
        south = True
    else:
        south = False

    # use PROJ string, assuming a default WGS84
    # crs = CRS.from_string(f'+proj=utm +zone=zone +south')

    # use dictionary to define the PROJ
    crs = CRS.from_dict({'proj': 'utm', 'zone': zone, 'south': south})

    return crs


def lonlat_to_xy(lon, lat, crs):
    p1 = Proj(crs, preserve_units=True)
    (y, x) = p1(lon, lat)
    return x, y


def xy_to_lonlat(x, y, crs):
    p1 = Proj(crs, preserve_units=True)
    (lon, lat) = p1(y, x, inverse=True)
    return lon, lat


def get_var(ncfile, varname):
    nc_ds = xr.open_dataset(ncfile, decode_coords="all")

    var = nc_ds[varname]

    var = var.transpose('time', 'y', 'x')

    var = var.rio.set_spatial_dims(x_dim='x', y_dim='y')

    var.attrs['affine'] = var.rio.transform()

    var.attrs['bounds'] = var.rio.bounds()

    return var

def get_ds(ncfile):
    nc_ds = xr.open_dataset(ncfile, decode_coords="all")
    nc_ds = nc_ds.transpose('time', 'y', 'x', ...)
    nc_ds = nc_ds.rio.set_spatial_dims(x_dim='x', y_dim='y')
    nc_ds.attrs['affine'] = nc_ds.snow_depth.rio.transform()
    nc_ds.attrs['bounds'] = nc_ds.snow_depth.rio.bounds()
    return nc_ds


def get_snow_depth_deltacrt(ncfile, varname):
    nc_ds = xr.open_dataset(ncfile, decode_coords="all")



def get_affine_bbox(nc_ds):
    """Get the bounding box of the nc variable
    arguments: nc dataset
    retrurns: affine
              bbox [min_lon,min_lat, max_lon, max_lat]
    """
    affine = nc_ds.rio.transform()
    bounds = nc_ds.rio.bounds()
    return affine, bounds


def bbox2polygon(bbox):
    """Convert bbox [min_lon,min_lat,max_lon, max_lat] to shapely polygon.
    arguments: bbox: [min_lon,min_lat, max_lon, max_lat], example [-113.2, 43.1, -113.1, 43.2]
    returns: polygon
    """
    p1 = geometry.Point(bbox[0], bbox[1])
    p2 = geometry.Point(bbox[2], bbox[1])
    p3 = geometry.Point(bbox[2], bbox[3])
    p4 = geometry.Point(bbox[0], bbox[3])
    pointlist = [p1, p2, p3, p4]
    poly = geometry.Polygon(pointlist)

    return poly


def write_geojson(bbox, outfile):
    poly = bbox2polygon(bbox)
    data ={"name":["The Caribou-Poker Creek and Poker Flats Research Watershed"], "geometry":[poly]}
    gdf = gpd.GeoDataFrame(data, crs='EPSG:4326')
    gdf.to_file(outfile, driver='GeoJSON')
    return outfile

def polygon_via_point(lon: float, lat: float, resolution: float):
    crs = get_crs(lon, lat)
    y, x, zone, band = utm.from_latlon(lat, lon)

    min_x = x - resolution/2
    min_y = y - resolution/2
    max_x = x + resolution/2
    max_y = y + resolution/2

    min_lon, min_lat = xy_to_lonlat(min_x, min_y, crs)
    max_lon, max_lat = xy_to_lonlat(max_x, max_y, crs)

    bbox = [min_lon, min_lat, max_lon, max_lat]

    return bbox2polygon(bbox)


def bbox_via_point(lon: float, lat: float, resolution: float):
    crs = get_crs(lon, lat)
    y, x, zone, band = utm.from_latlon(lat, lon)

    min_x = x - resolution/2
    min_y = y - resolution/2
    max_x = x + resolution/2
    max_y = y + resolution/2

    min_lon, min_lat = xy_to_lonlat(min_x, min_y, crs)
    max_lon, max_lat = xy_to_lonlat(max_x, max_y, crs)

    bbox = [min_lon, min_lat, max_lon, max_lat]

    return bbox


def to_percent(y, position):
    # Ignore the passed in position. This has the effect of scaling the default
    # tick locations.
    s = str(100 * y)

    # The percent symbol needs escaping in latex
    if matplotlib.rcParams['text.usetex'] is True:
        return s + r'$\%$'
    else:
        return s + '%'


def zonal_analyses(nc_var, daterange: list = None, times: list = None, poly: geometry.Polygon = None):
    # load and read shp-file with geopandas
    # shp_fo = r'../path/to/shp_file.shp'

    # shp_df = gpd.read_file(str(shapefile))

    # load and read netCDF-file to dataset and get datarray for variable
    # nc_fo = r'../path/to/netCDF_file.nc'
    # nc_ds = xr.open_dataset(nc_fo)

    # nc_var = nc_var.transpose('time', 'y', 'x')

    # get affine and bbox
    affine = nc_var.attrs['affine']
    bbox = list(nc_var.attris['bounds'])

    # decide if poly is not within bbox, return None for yes
    if not bbox2polygon(bbox).contains(poly):
        return None

    # select date range
    if daterange:
        nc_var = nc_var.sel(time=slice(*daterange))

    # get all years for which we have data in nc-file
    if times:
        times = times
    else:
        times = nc_var['time'].values

    # define output list to hold calculated statistic
    output = []

    # go over each time
    for time in times:
        # get values of variable pear year
        nc_arr = nc_var.sel(time=time, method='nearest')
        nc_arr_vals = nc_arr.values
        rst = rstats.zonal_stats(poly, nc_arr_vals, affine=affine, stats="mean min max")

        output.append({'time': time, 'geometry': poly, 'stat': rst})

        print({'time': time, 'geometry': poly, 'stat': rst})

    # convert output to geoDataFrame

    parsed_output = [[item['time'], item['stat'], item['geometry']] for item in output]

    gdf = gpd.GeoDataFrame(data=parsed_output, columns=['time', 'stat', 'geometry'])

    return gdf


def close_snex_time(nc_arr, time_ct, timethresh=12):
    nc_arr_range = nc_arr.copy(deep=True)

    # select times of days close to the time of t_ct
    # if len(nc_arr.time) >= 1:
    if nc_arr.time.shape[0] == 0:
        return None
    time_arr = np.array(nc_arr.time)
    time_arr = [pd.Timestamp(time).to_pydatetime() for time in time_arr]
    num = len(time_arr)
    idx = np.zeros(num, dtype=int)
    idx[:] = -1
    for i in range(num):
        # if abs(time_ct - time_arr[i]).seconds <= timethresh * 3600:
        time_ct_second_of_day = time_ct.hour * 3600 + time_ct.minute * 60 + time_ct.second
        time_i = time_arr[i]
        time_i_second_of_day = time_i.hour * 3600 + time_i.minute * 60 + time_i.second

        if abs(time_ct_second_of_day - time_i_second_of_day) <= timethresh * 3600:
            idx[i] = i
    # eliminate the wrong index
    idx = idx[idx >= 0]
    if idx.size == 0:
        return None
    # idx is the array including the index of times we pick up in the nc_arr
    pick_times = np.array(nc_arr.time[idx.tolist()])
    pick_times = [pd.Timestamp(time).to_pydatetime() for time in pick_times]
    nc_arr = nc_arr.sel(time=pick_times)
    return nc_arr


def close_s1_time(snex, time_ct, timethresh=12):
    # snex1 = snex.copy(deep=True)
    # select times of days close to the time of t_ct
    # if len(nc_arr.time) >= 1:
    # if len(snex) == 1:
    #    return snex
    time_arr = np.array(snex['time'])
    # time_arr = [pd.Timestamp(time).to_pydatetime() for time in time_arr]
    num = len(time_arr)
    idx = np.zeros(num, dtype=int)
    idx[:] = -1
    for i in range(num):
        time_ct_second_of_day = time_ct.hour*3600 + time_ct.minute*60 + time_ct.second
        time_i = pd.Timestamp(time_arr[i]).to_pydatetime()
        time_i_second_of_day = time_i.hour*3600 + time_i.minute*60 + time_i.second

        if abs(time_ct_second_of_day - time_i_second_of_day) <= timethresh * 3600:
            idx[i] = i
    # eliminate the wrong index
    idx = idx[idx >= 0]
    # if idx.size == 0:
    #    return None
    # idx is the array including the index of times we pick up in the nc_arr
    #pick_times = np.array(snex['time'].iloc[idx.tolist()])
    #pick_times = [pd.Timestamp(time).to_pydatetime() for time in pick_times]
    #nc_arr = snex[snex['time'](time=pick_times)

    return snex.iloc[idx]


def shape_by_xy_array(x, y):
    """ construct the shapely point or linestring
    :param x: array
    :param y: array
    :return:
    """

    num = x.shape[0]

    points = [[x[i], y[i]] for i in range(num)]

    if len(points) == 1:
        return geometry.Point(points)
    else:
        return geometry.LineString(points)

def zonal_analyses_based_snex(nc_var: xr.DataArray, snex: gpd.GeoDataFrame, res: float = None,
                        timeint: int = 24, timethresh: int = 6, method: str = 'average'):
    """zonal analyses of nc_var based on time and point within snex
    arguments:
        nc_var: sentinel1 snow depth data
        snex: observation data
        timeint: for example, 24, 48, 72, 84, etc.
        timethresh: time of day close to time of day of observation.
        method: average, nearest, close

    returns: combined GeoDataFrame, it includes the zonal statistic and observation results
    """
    # get affine and bbox
    affine = nc_var.attrs['affine']
    bbox = list(nc_var.attrs['bounds'])
    poly_bbox = bbox2polygon(bbox)

    # define the list to hold the calculated statistic
    output = []

    # go over each observation time
    total_num = nc_var.values.shape[0]
    count_fund = 0
    count_nan = 0
    for index, item in snex.iterrows():
        t_ct = item['time'].to_pydatetime()
        # convert to datetime
        # t_ct = datetime.strptime(time, '%Y-%m-%dT%H:%M')
        # covert boise local standard time to utc
        #t_ct = t_ct + timedelta(hours=6)
        time = datetime.strftime(t_ct, '%Y-%m-%dT%H:%M')
        # time range
        t_st = t_ct - timedelta(hours=timeint)
        t_ed = t_ct + timedelta(hours=timeint)

        nc_timerange = nc_var.sel(time=slice(t_st, t_ed))
        if nc_timerange.shape[0] == 0:
            continue

        try:
            # get values of variable pear year
            if method == 'average':
                nc_arr = nc_timerange

            elif method == 'nearest':

                # nearest time within time interval
                nc_arr = nc_timerange.sel(time=time, method='nearest', tolerance=f'{timeint}h')
                # nc_arr = nc_timerange.sel(time=time, method='nearest', tolerance=f'{timeint}h')
            elif method == 'close':
                nc_arr = close_snex_time(nc_arr=nc_timerange, time_ct=t_ct, timethresh=timethresh)
            else:
                pass
            # check if nc_arr is None
            if nc_arr is None:
                continue
            # check if nc_arr is empty
            nc_arr_vals = nc_arr.values
            if nc_arr_vals.size == 0:
                continue
            # do spatial clip
            poly = polygon_via_point(lon=item['Longitude'], lat=item['Latitude'], resolution=res)
            poly = geometry.box(*poly.bounds)
            # decide if poly is not within bbox, return None for yes
            # if not poly_bbox.contains(poly):  # should use poly_bbox.intersects() instead contain ?
            if not poly_bbox.intersects(poly):
                continue

            # clip by poly
            jsstr = to_geojson(poly)
            geometries = [json.loads(jsstr)]
            clipped = nc_arr.rio.clip(geometries, all_touched=True)
            x = np.array(clipped.x)
            y = np.array(clipped.y)
            xx, yy = np.meshgrid(x, y)
            xx = xx.flatten()
            yy = yy.flatten()
            clipped_geometry = shape_by_xy_array(xx, yy)
            values = clipped.values
            count_fund = count_fund + 1

            if len(list(nc_arr.time.shape)) == 0:
                number_time = 1
            else:
                number_time = nc_arr.time.shape[0]

            rst1 = {'mean': np.nanmean(values), 'min': np.nanmin(values), 'max': np.nanmax(values),
                    'std': np.nanstd(values), 'count': np.count_nonzero(~np.isnan(values)),
                    'num': number_time}

            if np.isnan(rst1['mean']):
                # why are there many clipped.values with np.nan value ?
                count_nan = count_nan + 1
                print(f'none at {t_ct}')
            # rst2 = rstats.zonal_stats(poly, nc_arr_vals, affine=affine, nodata=np.nan, stats="mean min max")[0]
            rst = rst1

            output.append({'location': item['Location'], 'time': time, 'geometry': poly, 's1geometry': clipped_geometry,
                           'mean': rst['mean'], 'min': rst['min'], 'max': rst['max'], 'std': rst['std'],
                           'count': rst['count'], 'num': rst['num'], 'snex_swe': item['SWE']*0.001,
                           'snex_sd': item['snow_depth'], 'hs': item['HS']*0.01})

            print({'time': time, 'mean': rst['mean'], 'obs': item['snow_depth']})

        except:
            continue

    # return empty GeoDataFrame for an empty list output
    if not output:
        return gpd.GeoDataFrame(columns=["foo", "bar", "geometry"], crs="EPSG:4326")

    # convert output to geoDataFrame
    parsed_output = [[item['location'], item['time'], item['mean'], item['min'], item['max'], item['std'],
                      item['count'], item['num'], item['snex_swe'], item['snex_sd'], item['hs'],
                      item['geometry'], item['s1geometry']] for item in output]

    gdf = gpd.GeoDataFrame(data=parsed_output, columns=['location', 'time', 'mean', 'min', 'max', 'std', 'count',
                                                        'num', 'snex_swe', 'snex_sd', 'hs', 'geometry', 's1geometry'])

    gdf = gdf.set_crs('epsg:4326')

    print(f'number of data with NAN is {count_nan}, number of data found are {count_fund}, '
          f'number of total data are {total_num}')

    return gdf


def one_to_one(nc_arr_one_time, snex_one_time, poly_bbox, res):
    poly = polygon_via_point(lon=snex_one_time['Longitude'], lat=snex_one_time['Latitude'], resolution=res)
    poly = geometry.box(*poly.bounds)
    # decide if poly is not within bbox, return None for yes
    # if not poly_bbox.contains(poly):  # should use poly_bbox.intersects() instead contain ?
    if not poly_bbox.intersects(poly):
        return None

    # clip by poly
    jsstr = to_geojson(poly)
    geometries = [json.loads(jsstr)]
    clipped = nc_arr_one_time.rio.clip(geometries, all_touched=True)
    x = np.array(clipped.x)
    y = np.array(clipped.y)
    xx, yy = np.meshgrid(x, y)
    xx = xx.flatten()
    yy = yy.flatten()
    clipped_geometry = shape_by_xy_array(xx, yy)
    values = clipped.values

    # rst = {'mean': np.nanmean(values), 'min': np.nanmin(values), 'max': np.nanmax(values),
    #                'std': np.nanmax(values), 'count': np.count_nonzero(~np.isnan(values))
    #        }

    rst = {'location': snex_one_time['Location'], 'time': snex_one_time['Time'],
           'geometry': snex_one_time['geometry'], 's1geometry': clipped_geometry, 'mean': np.nanmean(values),
           'min': np.nanmin(values), 'max': np.nanmax(values), 'std': np.nanstd(values),
           'count': np.count_nonzero(~np.isnan(values)), 'num': 1, 'snex_swe': snex_one_time['SWE'] * 0.001,
           'snex_sd': snex_one_time['snow_depth'], 'hs': snex_one_time['HS'] * 0.01}

    return rst


def zonal_analyses_based_sentinel1(nc_var: xr.DataArray, snex: gpd.GeoDataFrame, res: float = None,
                        timeint: int = 24, timethresh: int = 6, method: str = 'average'):
    """zonal analyses of nc_var based on time and point within snex
    arguments:
        nc_var: sentinel1 snow depth data
        snex: observation data
        timeint: for example, 24, 48, 72, 84, etc.
        timethresh: time of day close to time of day of observation.
        method: average, nearest, close.
    returns: combined GeoDataFrame, it includes the zonal statistic and observation results
    """
    # convert the Time column to the datetime column
    snex['time'] = pd.to_datetime(snex['Time'])

    # get affine and bbox
    affine = nc_var.attrs['affine']
    bbox = list(nc_var.attrs['bounds'])
    poly_bbox = bbox2polygon(bbox)

    # define the list to hold the calculated statistic
    output = []

    # go over each sentinel1 data
    total_num = nc_var.values.shape[0]
    count_fund = 0
    count_nan = 0
    for i in range(len(nc_var)):
        item = nc_var[i]
        time = pd.Timestamp(item.time.values).to_pydatetime()
        # covert boise local standard time to utc
        time = time + timedelta(hours=6)
        # time = datetime.strftime(t_ct, '%Y-%m-%dT%H:%M')
        # time range
        t_st = time - timedelta(hours=timeint)
        t_ed = time + timedelta(hours=timeint)

        snex_arr = snex[(snex['time'] >= np.datetime64(t_st)) & (snex['time'] <= np.datetime64(t_ed))]
        # check empty arr
        if snex_arr.empty:
            continue
        # three methods: average, nearest, close

        snex_tmp = snex_arr.copy(deep=True)

        if 'average' in method:
            count_fund = count_fund + snex_tmp.shape[0]
            try:
                # go over every time of snex data
                for index, snex_item in snex_tmp.iterrows():
                    rst = one_to_one(item, snex_item, poly_bbox, res)
                    if np.isnan(rst['mean']):
                        # why are there many clipped.values with np.nan value ?
                        count_nan = count_nan + 1
                        print(f'none at {time}')

                    output.append(rst)
            except:
                continue

        elif 'nearest' in method:
            try:
                snex2 = snex_tmp.set_index('time', inplace=False, drop=False)
                # get the index of nearest to time
                i = np.argmin(np.abs(snex2.index - time))
                count_fund = count_fund + 1
                rst = one_to_one(item, snex2.iloc[i], poly_bbox, res)
                output.append(rst)
                if np.isnan(rst['mean']):
                    count_nan = count_nan + 1
            except:
                continue

        elif 'close' in method:
            snex_tmp = close_s1_time(snex=snex_tmp, time_ct=time, timethresh=timethresh)
            count_fund = count_fund + len(snex_arr)
            for index, snex_item in snex_tmp.iterrows():
                rst = one_to_one(item, snex_item, poly_bbox, res)
                if np.isnan(rst['mean']):
                    # why are there many clipped.values with np.nan value ?
                    count_nan = count_nan + 1
                    print(f'none at {time}')

                output.append(rst)
        else:
            pass

    # return empty GeoDataFrame for an empty list output
    # if not output:
    #    return gpd.GeoDataFrame(columns=["foo", "bar", "geometry"], crs="EPSG:4326")

    # convert output to geoDataFrame
    parsed_output = [[item['location'], item['time'], item['mean'], item['min'], item['max'], item['std'],
                      item['count'], item['num'], item['snex_swe'], item['snex_sd'], item['hs'],
                      item['geometry'], item['s1geometry']] for item in output]

    gdf = gpd.GeoDataFrame(data=parsed_output, columns=['location', 'time', 'mean', 'min', 'max', 'std', 'count',
                                                        'num', 'snex_swe', 'snex_sd', 'hs', 'geometry', 's1geometry'])

    gdf = gdf.set_crs('epsg:4326')

    print(f'number of data with NAN is {count_nan}, number of data found are {count_fund}, '
          f'number of total data are {total_num}')

    return gdf

def one_s1_to_one_snotel(nc_arr_one_time, snex_one_time, poly_bbox, res):
    poly = polygon_via_point(lon=snex_one_time['Longitude'], lat=snex_one_time['Latitude'], resolution=res)
    poly = geometry.box(*poly.bounds)
    # decide if poly is not within bbox, return None for yes
    # if not poly_bbox.contains(poly):  # should use poly_bbox.intersects() instead contain ?
    if not poly_bbox.intersects(poly):
        return None

    # clip by poly
    jsstr = to_geojson(poly)
    geometries = [json.loads(jsstr)]
    clipped = nc_arr_one_time.rio.clip(geometries, all_touched=True)
    x = np.array(clipped.x)
    y = np.array(clipped.y)
    xx, yy = np.meshgrid(x, y)
    xx = xx.flatten()
    yy = yy.flatten()
    clipped_geometry = shape_by_xy_array(xx, yy)
    values = clipped.values

    # rst = {'mean': np.nanmean(values), 'min': np.nanmin(values), 'max': np.nanmax(values),
    #                'std': np.nanmax(values), 'count': np.count_nonzero(~np.isnan(values))
    #        }

    rst = {'time': snex_one_time['time'],
           'geometry': snex_one_time['geometry'], 's1geometry': clipped_geometry, 'mean': np.nanmean(values),
           'min': np.nanmin(values), 'max': np.nanmax(values), 'std': np.nanstd(values),
           'count': np.count_nonzero(~np.isnan(values)), 'num': 1, 'snex_sd': snex_one_time['snow_depth']}

    return rst

def one_s1_ds_to_one_snotel(nc_ds_one_time, snex_one_time, poly_bbox, res):
    poly = polygon_via_point(lon=snex_one_time['Longitude'], lat=snex_one_time['Latitude'], resolution=res)
    poly = geometry.box(*poly.bounds)
    # decide if poly is not within bbox, return None for yes
    # if not poly_bbox.contains(poly):  # should use poly_bbox.intersects() instead contain ?
    if not poly_bbox.intersects(poly):
        return None

    # clip by poly
    jsstr = to_geojson(poly)
    geometries = [json.loads(jsstr)]
    clipped = nc_ds_one_time.rio.clip(geometries, all_touched=True)
    x = np.array(clipped.x)
    y = np.array(clipped.y)
    xx, yy = np.meshgrid(x, y)
    xx = xx.flatten()
    yy = yy.flatten()
    clipped_geometry = shape_by_xy_array(xx, yy)
    val_sd = clipped.snow_depth.values
    val_cr = clipped.deltaCRt.values

    rst = {'time': snex_one_time['time'],
           'geometry': snex_one_time['geometry'], 's1geometry': clipped_geometry,
           'mean': np.nanmean(val_sd), 'min': np.nanmin(val_sd), 'max': np.nanmax(val_sd),'std': np.nanstd(val_sd),
           'meancr': np.nanmean(val_cr), 'mincr': np.nanmin(val_cr), 'maxcr': np.nanmax(val_cr),
           'stdcr': np.nanstd(val_cr),
           'count': np.count_nonzero(~np.isnan(val_sd)), 'num': 1, 'snex_sd': snex_one_time['snow_depth']}

    return rst


# def zonal_analyses_s1_snotel_based_s1

def zonal_analyses_s1_snotel_based_s1(nc_var: xr.DataArray, snex: gpd.GeoDataFrame, res: float = None,
                                   timeint: int = 24, timethresh: int = 6, method: str = 'average'):
    """zonal analyses of nc_var based on time and point within snex
    arguments:
        nc_var: sentinel1 snow depth data
        snex: observation data
        timeint: for example, 24, 48, 72, 84, etc.
        timethresh: time of day close to time of day of observation.
        method: average, nearest, close.
    returns: combined GeoDataFrame, it includes the zonal statistic and observation results
    """
    # get affine and bbox
    affine = nc_var.attrs['affine']
    bbox = list(nc_var.attrs['bounds'])
    poly_bbox = bbox2polygon(bbox)

    # define the list to hold the calculated statistic
    output = []

    # go over each sentinel1 data
    total_num = nc_var.values.shape[0]
    count_fund = 0
    count_nan = 0
    for i in range(len(nc_var)):
        item = nc_var[i]
        time = pd.Timestamp(item.time.values).to_pydatetime()
        # both s1 and SNOTEL time are in utc
        t_st = time - timedelta(hours=timeint)
        t_ed = time + timedelta(hours=timeint)
        snex_arr = snex[(snex['time'] >= np.datetime64(t_st)) & (snex['time'] <= np.datetime64(t_ed))]
        # check empty arr
        if snex_arr.empty:
            continue
        # three methods: average, nearest, close

        snex_tmp = snex_arr.copy(deep=True)

        if 'average' in method:
            count_fund = count_fund + snex_tmp.shape[0]
            try:
                # go over every time of snex data
                for index, snex_item in snex_tmp.iterrows():
                    rst = one_s1_to_one_snotel(item, snex_item, poly_bbox, res)
                    if np.isnan(rst['mean']):
                        # why are there many clipped.values with np.nan value ?
                        count_nan = count_nan + 1
                        print(f'none at {time}')

                    output.append(rst)
            except:
                continue

        elif 'nearest' in method:
            try:
                snex2 = snex_tmp.set_index('time', inplace=False, drop=False)
                # get the index of nearest to time
                i = np.argmin(np.abs(snex2.index - time))
                count_fund = count_fund + 1
                rst = one_to_one(item, snex2.iloc[i], poly_bbox, res)
                output.append(rst)
                if np.isnan(rst['mean']):
                    count_nan = count_nan + 1
            except:
                continue

        elif 'close' in method:
            snex_tmp = close_s1_time(snex=snex_tmp, time_ct=time, timethresh=timethresh)
            count_fund = count_fund + len(snex_arr)
            for index, snex_item in snex_tmp.iterrows():
                rst = one_to_one(item, snex_item, poly_bbox, res)
                if np.isnan(rst['mean']):
                    # why are there many clipped.values with np.nan value ?
                    count_nan = count_nan + 1
                    print(f'none at {time}')

                output.append(rst)
        else:
            pass

    # return empty GeoDataFrame for an empty list output
    # if not output:
    #    return gpd.GeoDataFrame(columns=["foo", "bar", "geometry"], crs="EPSG:4326")

    # convert output to geoDataFrame
    parsed_output = [[item['time'], item['mean'], item['min'], item['max'], item['std'],
                      item['count'], item['num'], item['snex_sd'],
                      item['geometry'], item['s1geometry']] for item in output]

    gdf = gpd.GeoDataFrame(data=parsed_output, columns=['time', 'mean', 'min', 'max', 'std',
                                                        'count', 'num', 'snex_sd',
                                                        'geometry', 's1geometry'])

    gdf = gdf.set_crs('epsg:4326')

    print(f'number of data with NAN is {count_nan}, number of data found are {count_fund}, '
          f'number of total data are {total_num}')

    return gdf

def zonal_analyses_s1_snotel_dry_wet_based_s1(nc_var: xr.Dataset, snex: gpd.GeoDataFrame, res: float = None,
                                   timeint: int = 24, timethresh: int = 6, method: str = 'average'):
    """zonal analyses of nc_var based on time and point within snex
    arguments:
        nc_var: sentinel1 dataset
        snex: observation data
        timeint: for example, 24, 48, 72, 84, etc.
        timethresh: time of day close to time of day of observation.
        method: average, nearest, close.
    returns: combined GeoDataFrame, it includes the zonal statistic and observation results
    """
    # get affine and bbox
    affine = nc_var.attrs['affine']
    bbox = list(nc_var.attrs['bounds'])
    poly_bbox = bbox2polygon(bbox)

    # define the list to hold the calculated statistic
    output = []

    # go over each sentinel1 data
    total_num = nc_var.time.values.shape[0]
    count_fund = 0
    count_nan = 0
    for i in range(len(nc_var.time)):
        item = nc_var.sel(time=nc_var.time[i])
        time = pd.Timestamp(item.time.values).to_pydatetime()
        # both s1 and SNOTEL time are in utc
        t_st = time - timedelta(hours=timeint)
        t_ed = time + timedelta(hours=timeint)
        snex_arr = snex[(snex['time'] >= np.datetime64(t_st)) & (snex['time'] <= np.datetime64(t_ed))]
        # check empty arr
        if snex_arr.empty:
            continue
        # three methods: average, nearest, close

        snex_tmp = snex_arr.copy(deep=True)

        if 'average' in method:
            count_fund = count_fund + snex_tmp.shape[0]
            try:
                # go over every time of snex data
                for index, snex_item in snex_tmp.iterrows():
                    rst = one_s1_ds_to_one_snotel(item, snex_item, poly_bbox, res)
                    if np.isnan(rst['mean']):
                        # why are there many clipped.values with np.nan value ?
                        count_nan = count_nan + 1
                        print(f'none at {time}')

                    output.append(rst)
            except:
                continue

        elif 'nearest' in method:
            try:
                snex2 = snex_tmp.set_index('time', inplace=False, drop=False)
                # get the index of nearest to time
                i = np.argmin(np.abs(snex2.index - time))
                count_fund = count_fund + 1
                rst = one_to_one(item, snex2.iloc[i], poly_bbox, res)
                output.append(rst)
                if np.isnan(rst['mean']):
                    count_nan = count_nan + 1
            except:
                continue

        elif 'close' in method:
            snex_tmp = close_s1_time(snex=snex_tmp, time_ct=time, timethresh=timethresh)
            count_fund = count_fund + len(snex_arr)
            for index, snex_item in snex_tmp.iterrows():
                rst = one_to_one(item, snex_item, poly_bbox, res)
                if np.isnan(rst['mean']):
                    # why are there many clipped.values with np.nan value ?
                    count_nan = count_nan + 1
                    print(f'none at {time}')

                output.append(rst)
        else:
            pass

    # return empty GeoDataFrame for an empty list output
    # if not output:
    #    return gpd.GeoDataFrame(columns=["foo", "bar", "geometry"], crs="EPSG:4326")

    # convert output to geoDataFrame
    parsed_output = [[item['time'], item['mean'], item['min'], item['max'], item['std'],
                      item['meancr'], item['mincr'], item['maxcr'], item['stdcr'],
                      item['count'], item['num'], item['snex_sd'],
                      item['geometry'], item['s1geometry']] for item in output]

    gdf = gpd.GeoDataFrame(data=parsed_output, columns=['time', 'mean', 'min', 'max', 'std',
                                                        'meancr', 'mincr', 'maxcr', 'stdcr', 'count',
                                                        'num', 'snex_sd', 'geometry',
                                                        's1geometry'])

    gdf = gdf.set_crs('epsg:4326')

    print(f'number of data with NAN is {count_nan}, number of data found are {count_fund}, '
          f'number of total data are {total_num}')

    return gdf



def draw_barplot(x, y):

    plt.bar(x, y)


def draw_line_plot(x, data_dic):
    fig, ax = plt.subplots()
    ax.plot(x, data_dic['min'], label='min', color='black')
    ax.plot(x, data_dic['mean'], label='mean', color='blue')
    ax.plot(x, data_dic['max'], label='max', color='red')
    ax.plot(x, data_dic['obs'], label='obs SWE', color='green')
    ax.legend()
    ax.set_title(f'Zonal Statistic of snow depth for {data_dic['location']}')
    ax.set_xlabel("Observation Date, date")
    ax.set_ylabel("snow depth, m")
    plt.show()


def draw_combined_gdf(gdf, res, timeint, timethresh, base, method, outfile, precip):

    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]

    corr = gdf['mean'].corr(gdf['hs'])

    # draw gdf on the plot

    minv = np.array(gdf['min'])

    meanv = np.array(gdf['mean'])

    maxv = np.array(gdf['max'])

    stderror = np.array(gdf['std'])

    times = np.array(gdf['time'])

    obs = np.array(gdf['snex_swe'])

    # times1 = [datetime.strptime(item, '%Y-%m-%dT%H:%M') for item in times]

    times1 = [pd.Timestamp(item) for item in times]

    days = [(item - times1[0]).total_seconds()/(24*3600.0) for item in times1]

    # convert times to datetime

    xtks = np.arange(days[0], np.round(days[-1]) + 2, 5)

    xtkslb = [str(int(item)) for item in xtks]

    start_time = times1[0].strftime("%Y%m%d")

    loc = gdf['location'].iloc[0]

    data_dic = {'location': loc, 'min': minv, 'mean': meanv, 'max': maxv, 'obs': obs}

    fig, ax = plt.subplots()

    fig.set_figwidth(30)

    # ax.plot(days, data_dic['min'], label='min', color='black')
    # ax.plot(days, data_dic['mean'], label='mean', color='blue')
    ax.plot(days, gdf['mean'], label='mean', color='blue')

    ax.errorbar(days, gdf['mean'], yerr=stderror, color='blue')
    
    # ax.plot(days, data_dic['max'], label='max', color='red')

    # ax.plot(days, gdf['snex_swe'], label='obs SWE', color='green')

    ax.plot(days, gdf['snex_sd'], label='obs snow depth', color='red')

    # ax.plot(days, gdf['hs'], label='obs snow height', color='black')

    ax.set_xticks(xtks, labels=xtkslb)

    ax.legend()

    corr = float("{:.2f}".format(corr))

    ax.text(0.8, 0.8, f'correlation: {corr}', color='red', transform=ax.transAxes)

    ax.set_title(f'{base}-based {method} Zonal Statistic of Snow Depth for {loc} with {str(int(res))} m and'
                  f' {int(timeint)} hours time interval with {int(timethresh)} hours threshold with precip {precip}')
    ax.set_xlabel(f'Observation day, days from staring {start_time}')
    ax.set_ylabel("snow depth, m")

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)

    plt.show()


def draw_diff_error_count_plot(gdf, res, timeint, timethresh, base, method, outfile, precip):
    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]

    corr = gdf['mean'].corr(gdf['hs'])

    # draw gdf on the plot

    minv = np.array(gdf['min'])

    meanv = np.array(gdf['mean'])

    maxv = np.array(gdf['max'])

    stderror = np.array(gdf['std'])

    times = np.array(gdf['time'])

    snow_depth_diff = gdf['hs'] - gdf['mean']

    # times1 = [datetime.strptime(item, '%Y-%m-%dT%H:%M') for item in times]

    times1 = [pd.Timestamp(item) for item in times]

    days = [(item - times1[0]).total_seconds() / (24 * 3600.0) for item in times1]

    # convert times to datetime

    xtks = np.arange(days[0], np.round(days[-1]) + 2, 5)

    xtkslb = [str(int(item)) for item in xtks]

    start_time = times1[0].strftime("%Y%m%d")

    loc = gdf['location'].iloc[0]

    # data_dic = {'location': loc, 'min': minv, 'mean': meanv, 'max': maxv, 'obs': obs}

    # fig, ax = plt.subplots()
    fig, [ax0, ax1,  ax2] = plt.subplots(nrows=3, ncols=1)
    fig.set_figwidth(30)
    fig.set_figheight(30)


    # ax.plot(days, data_dic['min'], label='min', color='black')
    # ax.plot(days, data_dic['mean'], label='mean', color='blue')
    # ax0.plot(days, gdf['snex_sd'] - gdf['mean'], label='snow depth difference', color='blue')

    # mean, obs
    ax0.plot(days, gdf['mean'], label='mean', color='blue')
    ax0.errorbar(days, gdf['mean'], yerr=stderror, color='blue')
    ax0.plot(days, gdf['snex_sd'], label='obs snow depth', color='red')
    ax0.set_xticks(xtks, labels=xtkslb)
    ax0.legend()

    corr = float("{:.2f}".format(corr))

    ax0.text(0.8, 0.8, f'correlation: {corr}', color='red', transform=ax0.transAxes)

    ax0.set_title(f'{base}-based {method} Zonal Statistic of Snow Depth for {loc} with {str(int(res))} m and'
                  f' {int(timeint)} hours time interval with {int(timethresh)} hours threshold with precip {precip}')
    ax0.set_xlabel(f'Observation day, days from staring {start_time}')
    ax0.set_ylabel("snow depth, m")

    # differnec with mean standard error bar
    ax1.errorbar(days, gdf['snex_sd'] - gdf['mean'], yerr=gdf['std'], label='Difference with mean standard error', color='red')

    ax1.set_xticks(xtks, labels=xtkslb)
    ax1.legend()

    corr = float("{:.2f}".format(corr))
    ax1.text(0.8, 0.8, f'correlation: {corr}', color='red', transform=ax1.transAxes)

    ax1.set_title(f'{base}-based {method} Zonal Statistic of Snow Depth for {loc} with {str(int(res))} m and'
                  f' {int(timeint)} hours time interval with {int(timethresh)} hours threshold with precip {precip}')

    ax1.set_xlabel(f'Observation day, days from staring {start_time}')

    ax1.set_ylabel("snow depth, m")

    # Number of pixels of mean
    ax2.plot(days, gdf['count'], label='Number of pixels', color='green')
    ax2.set_xticks(xtks, labels=xtkslb)
    ax2.set_xlabel(f'Observation day, days from staring {start_time}')
    ax2.legend()

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)

    plt.show()


def draw_diff_error_count_plot_index(gdf, res, timeint, timethresh, base, method, outfile, precip):
    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]
    gdf = gdf[(gdf['mean'] >= 0.0) & (gdf['hs'] >= 0.0)]
    corr = gdf['mean'].corr(gdf['hs'])

    # draw gdf on the plot

    minv = np.array(gdf['min'])

    meanv = np.array(gdf['mean'])

    maxv = np.array(gdf['max'])

    stderror = np.array(gdf['std'])

    loc = gdf['location'].iloc[0]

    fig, [ax0, ax1,  ax2] = plt.subplots(nrows=3, ncols=1)
    fig.set_figwidth(30)
    fig.set_figheight(30)

    # mean, obs
    s1_mean = "{:.2f}".format(gdf['mean'].mean())
    s1_std = "{:.2f}".format(gdf['mean'].std())
    obs_mean = "{:.2f}".format(gdf['snex_sd'].mean())
    obs_std = "{:.2f}".format(gdf['snex_sd'].std())

    ax0.plot(gdf.index, gdf['mean'], label=f'sentinel1 snow depth, mean {s1_mean},'
                                           f' std {s1_std} ', color='blue')
    ax0.errorbar(gdf.index, gdf['mean'], yerr=stderror, color='blue')
    ax0.plot(gdf.index, gdf['snex_sd'], label=f'snex snow depth, mean {obs_mean},'
                                              f'std {obs_std}', color='red')
    # ax0.set_xticks(xtks, labels=xtkslb)
    ax0.legend()

    corr1 = float("{:.2f}".format(corr))

    ax0.text(0.5, 0.9, f' Correlation {corr1}, Number of observations {len(gdf)}', color='red', transform=ax0.transAxes)

    ax0.set_title(f'{base}-based {method} Zonal Statistic of Snow Depth for {loc} with {str(int(res))} m and'
                  f' {int(timeint)} hours time interval with {int(timethresh)} hours threshold with precip {precip}')
    ax0.set_xlabel(f'Sequence of Observations')
    ax0.set_ylabel("snow depth, m")

    # differnec with mean standard error bar
    diff = gdf['mean'] - gdf['hs']
    #diff_mean = "{:.2f}".format(diff.mean())
    diff_mean = round(diff.mean(), 2)
    #diff_std = "{:.2f}".format(diff.std())
    diff_std = round(diff.std(), 2)
    ax1.errorbar(gdf.index, diff, yerr=stderror, label='Difference with mean standard error', color='red')

    #ax1.set_xticks(xtks, labels=xtkslb)
    ax1.legend()

    corr1 = float("{:.2f}".format(corr))

    ax1.text(0.6, 0.9, f'Correlation {corr1}, mean of difference {diff_mean}, std {diff_std}', color='red',
             transform = ax1.transAxes)

    ax1.set_title(f'The difference of Snow Depth Depth for {loc} with {str(int(res))} m and'
                  f' {int(timeint)} hours time interval with {int(timethresh)} hours threshold with precip {precip}')

    ax1.set_xlabel(f'Sequence of Observations')

    ax1.set_ylabel("snow depth, m")

    # Number of pixels of mean
    ax2.plot(gdf.index, gdf['count'], label='Number of pixels', color='green')
    # ax2.set_xticks(xtks, labels=xtkslb)
    ax2.set_xlabel(f'Sequence of Observations')
    ax2.legend()

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)

    plt.show()

def draw_s1_snotel_diff_error_count_plot(gdf, res, timeint, timethresh, base, method, outfile, precip, loc):
    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]
    gdf = gdf[(gdf['mean'] >= 0.0) & (gdf['snex_sd'] >= 0.0)]

    corr = gdf['mean'].corr(gdf['snex_sd'])

    ym = np.array(gdf['mean'])
    xm = np.array(gdf['snex_sd'])

    r = np.corrcoef(ym,xm)
    corr = r[0,1]
    rmse = np.sqrt(((xm - ym) ** 2).mean())
    mbe = np.mean(ym - xm)


    # draw gdf on the plot

    minv = np.array(gdf['min'])

    meanv = np.array(gdf['mean'])

    maxv = np.array(gdf['max'])

    stderror = np.array(gdf['std'])


    fig, [ax0, ax1,  ax2] = plt.subplots(nrows=3, ncols=1)
    fig.set_figwidth(30)
    fig.set_figheight(30)

    # mean, obs
    s1_mean = "{:.2f}".format(gdf['mean'].mean())
    s1_std = "{:.2f}".format(gdf['mean'].std())
    obs_mean = "{:.2f}".format(gdf['snex_sd'].mean())
    obs_std = "{:.2f}".format(gdf['snex_sd'].std())
    corr1 = float("{:.2f}".format(corr))
    rmse1 = "{:.2f}".format(rmse)
    mbe1 = "{:.2f}".format(mbe)


    x = gdf.index

    xtks = x

    xtklbs = gdf.time[x].dt.strftime('%m/%d')

    # get every 10th
    xtks = xtks[0::10]

    xtklbs = xtklbs[0::10]

    ax0.plot(x, gdf['mean'], label=f'S1 Snow Depth, mean {s1_mean},'
                                           f' std {s1_std} ', color='blue')
    ax0.errorbar(x, gdf['mean'], yerr=stderror, color='blue')

    ax0.plot(x, gdf['snex_sd'], label=f'SNOTEL Snow Depth, mean {obs_mean},'
                                              f'std {obs_std}', color='red')

    ax0.set_xticks(xtks, xtklbs)

    ax0.legend()

    ax0.text(0.3, 0.9, f'RMSE {rmse1}, R {corr1}, MBE {mbe1}, Number of observations {len(gdf)}', color='red', transform=ax0.transAxes)

    ax0.set_title(f'{base.capitalize()}-based Zonal Statistic of {method.capitalize()} S1 Snow Depth with {str(int(res))} m and'
                  f' {int(timeint)} Hours Time Interval and SNOTEL Snow Depth for {loc}')
    ax0.set_xlabel(f'Sequence of Observations')
    ax0.set_ylabel("snow depth, m")

    # differnec with mean standard error bar
    diff = gdf['mean'] - gdf['snex_sd']
    #diff_mean = "{:.2f}".format(diff.mean())
    diff_mean = round(diff.mean(), 2)
    #diff_std = "{:.2f}".format(diff.std())
    diff_std = round(diff.std(), 2)
    ax1.errorbar(x, diff, yerr=stderror, label='Difference with Mean Standard Error', color='red')

    ax1.set_xticks(xtks, xtklbs)
    ax1.legend()

    corr1 = float("{:.2f}".format(corr))

    ax1.text(0.6, 0.9, f'Correlation {corr1}, Mean of Difference {diff_mean}, STD {diff_std}', color='red',
             transform = ax1.transAxes)

    ax1.set_title(f'The Difference of S1 Snow Depth Depth with {str(int(res))} m and'
                  f' {int(timeint)} Hours Time Interval and SNOTEL Snow Depth for {loc}')

    ax1.set_xlabel(f'Sequence of Observations')

    ax1.set_ylabel("Snow Depth, m")

    # Number of pixels of mean
    ax2.plot(x, gdf['count'], label='Number of pixels', color='green')
    # ax2.set_xticks(xtks, labels=xtkslb)
    ax2.set_xlabel(f'Sequence of Observations')
    ax2.set_xticks(xtks, xtklbs)
    ax2.legend()

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)

    plt.show()


def draw_s1_sd_cr_snotel_plot(gdf, res, timeint, timethresh, base, method, outfile, precip, loc):
    # draw S1 snow depth, delataCR in ratio, and SNOTEL snow depth in the same figure
    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]
    gdf = gdf[(gdf['mean'] >= 0.0) & (gdf['snex_sd'] >= 0.0)]

    corr = gdf['mean'].corr(gdf['snex_sd'])

    ym = np.array(gdf['mean'])
    xm = np.array(gdf['snex_sd'])

    r = np.corrcoef(ym,xm)
    corr = r[0,1]
    rmse = np.sqrt(((xm - ym) ** 2).mean())
    mbe = np.mean(ym - xm)


    # draw gdf on the plot

    minv = np.array(gdf['min'])

    meanv = np.array(gdf['mean'])

    maxv = np.array(gdf['max'])

    stderror = np.array(gdf['std'])


    fig, [ax0, ax1,  ax2] = plt.subplots(nrows=3, ncols=1)
    fig.set_figwidth(30)
    fig.set_figheight(30)

    # xticks and xtick labels
    x = gdf.index
    xtks = x
    xtklbs = gdf.time[x].dt.strftime('%m/%d')
    # get every 10th
    xtks = xtks[0::10]
    xtklbs = xtklbs[0::10]

    # statistic
    obs_mean = "{:.2f}".format(gdf['snex_sd'].mean())
    obs_std = "{:.2f}".format(gdf['snex_sd'].std())
    corr1 = float("{:.2f}".format(corr))
    rmse1 = "{:.2f}".format(rmse)
    mbe1 = "{:.2f}".format(mbe)

    # ax0, draw s1 snow depth
    # s1 statistic
    s1_mean = "{:.2f}".format(gdf['mean'].mean())
    s1_std = "{:.2f}".format(gdf['mean'].std())


    ax0.plot(x, gdf['mean'], label=f'S1 Depth, mean {s1_mean},'
                                           f' std {s1_std} ', color='blue')
    # ax0.text(0.3, 0.9, f'Correlation between S1 and SNOTEL snow depths {corr1}', color='red')
    ax0.set_xticks(xtks, xtklbs)
    ax0.legend()
    ax0.text(0.1, 0.9, f'RMSE {rmse1}, R {corr1}, MBE {mbe1}, Number of observations {len(gdf)}', color='red', transform=ax0.transAxes)

    ax0.set_title(f'{base.capitalize()}-based Zonal Statistic of {method.capitalize()} S1 Snow Depth with {str(int(res))} m and'
                  f' {int(timeint)} Hours Time Interval for {loc}')
    ax0.set_xlabel(f'Sequence of Observations')
    ax0.set_ylabel("S1 Snow Depth")

    # draw delta CR in ratio

    # s1 deltacr statistic
    s1_cr_mean = "{:.2f}".format(gdf['mean'].mean())
    s1_cr_std = "{:.2f}".format(gdf['mean'].std())
    corr = gdf['meancr'].corr(gdf['snex_sd'])
    ax1.plot(x, gdf['meancr'], label=f'S1 DeltaCR, mean {s1_cr_mean},'
                                      f'std {s1_cr_std}', color='red')
    ax1.set_xticks(xtks, xtklbs)
    ax1.legend()
    corr1 = float("{:.2f}".format(corr))
    ax1.text(0.1, 0.9, f'Correlation between deltaCR and SNOTEL Snow Depth {corr1}', color='red', transform=ax1.transAxes)
    ax1.set_title(f'S1 deltaCR for {loc}')
    ax1.set_xlabel(f'Sequence of Observations')
    ax1.set_ylabel("Delta Cross Ratio")

    # draw SNOTEL snow depth
    corr = gdf['mean'].corr(gdf['snex_sd'])
    ax2.plot(x, gdf['snex_sd'], label=f'SNOTEL Snow Depth, mean {obs_mean},'
                                      f'std {obs_std}', color='red')
    ax2.set_xticks(xtks, xtklbs)
    ax2.legend()
    corr1 = float("{:.2f}".format(corr))
    ax2.text(0.1, 0.9, f'Correlation between S1 and SNOTEL snow depths {corr1}', color='red', transform=ax2.transAxes)
    ax2.set_title(f'SNOTEL Snow Depth for {loc}')
    ax2.set_xlabel(f'Sequence of Observations')
    ax2.set_ylabel("Snow Depth, m")
    ax2.legend()

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)

    plt.show()




def draw_histgram(gdf, res, timeint, timethresh, base, method, outfile, precip, loc):
    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]

    corr = gdf['mean'].corr(gdf['snex_sd'])

    # loc = gdf['location'].iloc[0]
    # draw gdf on the plot
    n_bins = 10
    colors = ['green', 'blue', 'lime', 'red']
    # fig, ax = plt.subplots()
    # fig.set_figwidth(30)
    fig, [(ax0, ax1), (ax2, ax3)] = plt.subplots(nrows=2, ncols=2)
    fig.set_figwidth(30)
    fig.set_figheight(30)

    fig.suptitle(f'{base.capitalize()}-based {method.capitalize()} Zonal Statistic of Snow Depth with {str(int(res))} m and'
                  f' {int(timeint)} Hours Time Interval for {loc}')

    diff = gdf['snex_sd'] - gdf['mean']

    # ax0.hist(diff, bins=n_bins, histtype='bar', color=colors[0], label=f'Total Count {len(gdf['snex_sd'])}')

    ax0.hist(diff, bins=n_bins, histtype='bar', weights=np.ones(len(diff)) / len(diff), label=f'Total Count {len(diff)}' )

    # Set the formatter
    ax0.yaxis.set_major_formatter(PercentFormatter(1))

    ax0.legend(prop={'size': 10})
    ax0.set_title('Difference of Snow Depth')
    ax0.set_xlabel(f'Difference of Snow Depth, m')
    ax0.set_ylabel("Count")


    ax1.hist(gdf['std'], n_bins, histtype='barstacked', stacked=True,  color=colors[1])
    ax1.set_title('Standard Error of Sentinel1 Snow Depth')
    ax1.set_xlabel(f'Snow Depth Standard Error, m')
    ax1.set_ylabel("Count")

    ax2.hist(gdf['count'], n_bins, histtype='barstacked', stacked=True,  color=colors[2])
    ax2.set_title('Number of Pixels for Average of Sentinel1 Snow Depth')
    ax2.set_xlabel(f'Number of Pixels')
    ax2.set_ylabel("Count")

    ax3.hist(gdf['num'], n_bins, histtype='barstacked', stacked=True,  color=colors[3])
    ax3.set_title('Number of Images for Average')
    ax3.set_xlabel(f'Number of images for Average of Sentinel1 Snow Depth')
    ax3.set_ylabel("Count")
    ax3.legend()
    num_bins = 10
    #n, bins, patches = ax.hist(gdf['snex_sd'] - gdf['mean'], num_bins,
    #                           density=1,
    #                           color='green',
    #                           alpha=0.7)

    #num = len(gdf['snex_sd'] - gdf['mean'])
    #ax.hist(gdf['snex_sd'] - gdf['mean'])
    #y = ((1 / (np.sqrt(2 * np.pi) * sigma)) *
    #     np.exp(-0.5 * (1 / sigma * (bins - mu)) ** 2))
    #ax.plot(bins, y, '--', color='black')

    #ax.legend()
    # ax.set_xlabel('X-Axis')
    # ax.set_ylabel('Y-Axis')

    #corr = float("{:.2f}".format(corr))

    #ax.text(0.8, 0.8, f'correlation: {corr}, total number of difference {num}', color='red', transform=ax.transAxes)


    #ax.set_title(f'Histgram of Difference of snex - sentinel1 Snow Depth with {str(int(res))} m and {int(timeint)} hours with {int(timethresh)} hours with precip'
    #             f' {precip} {base} based {method} Zonal Statistic of Snow Depth for {gdf['location'].iloc[0]}')
    #ax.set_xlabel(f'snow depth difference values, m')
    #ax.set_ylabel("Frequency")

    plt.show()

    # create directory if not exist
    Path(outfile).parent.mkdir(exist_ok=True)

    fig.savefig(outfile)

    plt.show()


def draw_scatter_plot(gdf, res, timeint):

    # get rid of rows that does not include valid mean

    gdf = gdf[~gdf['mean'].isnull()]

    corr = gdf['mean'].corr(gdf['snex_swe'])

    # draw gdf on the plot

    minv = np.array(gdf['min'])

    meanv = np.array(gdf['mean'])

    maxv = np.array(gdf['max'])

    times = np.array(gdf['time'])

    obs = np.array(gdf['snex_swe'])

    times1 = [datetime.strptime(item, '%Y-%m-%dT%H:%M') for item in times]

    days = [(item - times1[0]).total_seconds()/(24*3600.0) for item in times1]

    # convert times to datetime

    xtks = np.arange(days[0], np.round(days[-1]) + 2)

    # xtkslb = [(times1[0] + timedelta(item)).strftime("%m%d") for item in xtks]

    xtkslb = [str(int(item)) for item in xtks]

    start_time = times1[0].strftime("%Y%m%d")

    loc = gdf['location'].iloc[0]

    data_dic = {'location': loc, 'min': minv, 'mean': meanv, 'max': maxv, 'obs': obs}

    fig, ax = plt.subplots()
    # ax.plot(days, data_dic['min'], label='min', color='black')
    # ax.plot(days, data_dic['mean'], label='mean', color='blue')
    # ax.plot(days, data_dic['max'], label='max', color='red')
    # ax.plot(days, data_dic['obs'], label='obs SWE', color='green')

    ax.scatter(days, gdf['mean'], c='blue', label='mean',
               alpha=0.3, edgecolors='none')
    ax.scatter(days, gdf['snex_swe'], c='green', label='obs',
               alpha=0.3, edgecolors='none')

    ax.set_xticks(xtks, labels=xtkslb)
    ax.legend()

    corr = float("{:.2f}".format(corr))

    ax.text(0.8, 0.8, f'correlation: {corr}', color='red', transform=ax.transAxes)

    ax.set_title(f'{str(int(res))} m and time interval {timeint} Zonal Statistic of Snow Depth for {loc}')
    ax.set_xlabel(f'Observation day, days from staring {start_time}')
    ax.set_ylabel("snow depth, m")
    plt.show()


def draw_image(var, time: str):
    """
    Draw image at the time.
    :param var:
    :param time: for example '2021-03-21T13:33:18.000000000'.
    :return:
    """
    # draw the image at the time
    x, y = np.meshgrid(var.x, var.y)
    img1 = var.sel(time=time, method='nearest')
    plt.title(f'Snow Depth at Time {time[:10]}')
    plt.xlabel('Longitude')
    plt.ylabel('Latitude')
    mgplot = plt.imshow(img1, cmap=plt.cm.RdBu)

    plt.show()


def draw_bar_plot_at_location(var, loc):
    # draw vector at the loc, loc is a list [lon,lat]
    ts1 = var.sel(y=loc[1], x=loc[0], method='nearest')

    plt.bar(ts1.time, ts1.data)

    plt.show()


def subset_dataset_via_box(var, bbox):
    # select the interesting area (AOI) based on bbox [min_lon, max_lon, min_lat, max_lat]

    min_lon = bbox[0]
    max_lon = bbox[2]
    min_lat = bbox[1]
    max_lat = bbox[3]

    # cropped_ds = ds.sel(y=slice(min_lat,max_lat), x=slice(min_lon,max_lon))

    mask_lon = (var.x >= min_lon) & (var.x <= max_lon)

    mask_lat = (var.y >= min_lat) & (var.y <= max_lat)

    cropped_ds = var.where(mask_lon & mask_lat, drop=True)

    return cropped_ds


def zonal_analyze_shapefile(ncfile, varname, shpfile):
    """Do zonal analysis of the variable in the ncfile with zone defined by shpfile.
    :param ncfile:
    :param varname:
    :param shpfile:
    :return:
    """
    nc_ds = xr.open_dataset(ncfile)

    shp_df = gpd.read_file(str(shpfile))

    mpolys = geometry.MultiPolygon([poly for poly in shp_df['geometry']])

    nc_var = nc_ds[varname]

    nc_var = nc_var.transpose('time', 'y', 'x')

    nc_var = nc_var.rio.set_spatial_dims(x_dim='x', y_dim='y')

    for poly in mpolys.geoms:

        gdf_poly = zonal_analyses(nc_var=nc_var, poly=poly)

        minv = [item[0]['min'] for item in gdf_poly['stat']]

        meanv = [item[0]['mean'] for item in gdf_poly['stat']]

        maxv = [item[0]['max'] for item in gdf_poly['stat']]

        bounds = tuple([float("{:.2f}".format(x)) for x in poly.bounds])

        data_dic = {'bounds': bounds, 'min': minv, 'mean': meanv, 'max': maxv}

        draw_line_plot(gdf_poly['time'], data_dic)

    print('completed zonal_analyses...')


# read the variable


# bT = bT.rio.set_spatial_dims(x_dim='lon', y_dim='lat')

def vertor_analyze(ncfile, varname, loc):
    """analyze the vector at the loc
    :param ncfile:
    :param varname:
    :param loc:[lon, lat]
    :return:
    """
    nc_ds = xr.open_dataset(ncfile)

    var = nc_ds[varname]

    var = var.transpose('time', 'y', 'x')

    var = var.rio.set_spatial_dims(x_dim='x', y_dim='y')

    # draw bar plot
    # loc = [43.5, -113.2]

    draw_bar_plot_at_location(var, loc)


def zonal_analyze_bbox(ncfile, varname, bbox=None, daterange=None, times=None):
    """zonal analysis of the variable with zone defined by bbox.
    :param ncfile:
    :param varname:
    :param bbox: [min_lon,min_lat, max_lon, max_lat], example [-113.2, 43.1, -113.1, 43.2].
    :return:
    """
    # select an AOI via bbox
    nc_ds = xr.open_dataset(ncfile)

    var = nc_ds[varname]

    var = var.transpose('time', 'y', 'x')

    var = var.rio.set_spatial_dims(x_dim='x', y_dim='y')

    # bbox = [-113.1, -113.2, 43.1, 43.2]

    # sub_var = subset_dataset_via_box(var, bbox)

    poly = geometry.MultiPolygon([bbox2polygon(bbox)])

    gdf_poly = zonal_analyses(nc_var=var, poly=poly)

    minv = [item[0]['min'] for item in gdf_poly['stat']]

    meanv = [item[0]['mean'] for item in gdf_poly['stat']]

    maxv = [item[0]['max'] for item in gdf_poly['stat']]

    bounds = tuple([float("{:.2f}".format(x)) for x in poly.bounds])

    data_dic = {'bounds': bounds, 'min': minv, 'mean': meanv, 'max': maxv}

    draw_line_plot(gdf_poly['time'], data_dic)


def write_geotiff(ncfile):
    '''
    var = get_var(ncfile)

    sd.rio.crs

    lat= var.y

    lat_cent = float(lat[int(len(lat)/2 - 1)])

    lon = var.x

    lon_cent = float(lon[int(len(lon)/2 - 1)])

    # crs = get_epsg_code(lon_cent,lat_cent)

    crs = var.rio.estimate_utm_crs().to_epsg()

    # Define the CRS projection

    # sd.rio.write_crs("epsg:4326", inplace=True)

    # save the GeoTIFF file:

    # sd.rio.to_raster(r"/home/jiangzhu/data/crrel/spicy-test/snow_depth.tif")

    # save the NetCDF file:

    var.to_netcdf('/home/jiangzhu/data/crrel/spicy-test/snow_depth_by_xarray.nc')

    '''


# main program

def main():
    parser = ArgumentParser()
    parser.add_argument('--ncfile', type=str,  required=True)
    parser.add_argument('--varname', type=str, default='snow_depth')
    parser.add_argument('--shpfile', type=str, default=None)
    parser.add_argument('--bbox', nargs="*", type=float, default=None)
    parser.add_argument('--location', nargs="*", type=float, default=None)
    parser.add_argument('--time', type=str, default=None)
    parser.add_argument('--daterange', nargs="*", type=str, default=None)
    parser.add_argument('--workdir', type=str, default='.')  
    
    args = parser.parse_args()

    # example
    #ncfile='/home/jiangzhu/data/crrel/sentinel1_snowEx21_ts_snow_pits_v1/sentinel1.nc'
    #varname='snow_depth'
    #shpfile=None
    #bbox=[-116.6707976743838 43.136861475176104 -114.87764927713005 44.7157595218886]
    #location=[-115, 43.5]
    #daterange=[2020-11-15, 2021-11-30]
    #time='2021-03-21T13:33:18.000000000'

    # zonal analysis with shapefile
    if args.shpfile:
        zonal_analyze_shapefile(ncfile=args.ncfile, varname=args.varname, shpfile=args.shpfile)

    # vector analysis with location
    if args.location:
        vertor_analyze(args.ncfile, args.varname, args.location)

    # zonal analysis with bbox
    if args.bbox:
        zonal_analyze_bbox(args.ncfile, args.varname, args.bbox)

    # Draw plot at time
    if args.time:
        var = get_var(args.ncfile, args.varname)
        draw_image(var=var, time=args.time)

    print("completed ...")


if __name__ == '__main__':
    main()
